/* Table for storing sequences represented by an LZW code */
#include <hammer/hammer.h>
#include <hammer/glue.h>
// malloc, free
#include <stdlib.h>
// strlen
#include <string.h>

#include "lzw.h"

//FILE *debug; // DEBUG

#define BITLIMIT_9 (ctx->earlychange ? 511 : 512)
#define BITLIMIT_10 (ctx->earlychange ? 1023 : 1024)
#define BITLIMIT_11 (ctx->earlychange ? 2047 : 2048)
/*
 *  Since bit lengths larger than 12 aren't allowed, EarlyChange doesn't matter here.
 */
#define BITLIMIT_12 4096


static void
LZW_clear_table(LZW_context_T *ctx)
{
	/*
	 *  Optimizations: since we leave the entries 0-257 fixed or empty, we don't need to free() them explicitly.
	 *  And since codes are added to the table sequentially, we don't need to look past ctx->next;
	 */
	for(int i = 258; i < ctx->next; ++i)
	{
		HBytes * sequence = ctx->lzw_code_table[i];
		if(sequence != NULL)
		{
			/*
			 * Assumption: only the HBytes in the LZW table refer to the particular uint8_t arrays we're freeing.
			 */
			free((uint8_t *)sequence->token);
			free(sequence);
		}
		ctx->lzw_code_table[i] = NULL;
	}
}

/*
 * Creates a HBytes from an array of bytes and its length, and inserts it into the lzw dictionary in ctx.
 * Also increments ctx->next. The HBytes will keep the token pointer, to be freed later in lzw_clear_table or init_lzw_context.
 */
static void
lzw_table_insert(LZW_context_T *ctx, uint8_t *token, size_t token_len)
{
	HBytes * next_entry = malloc(sizeof(HBytes));
	next_entry->token = token;
	next_entry->len = token_len;
	ctx->lzw_code_table[ctx->next] = next_entry;
	ctx->next++;
}

HParser *p_lzwdata;
static LZW_context_T *context;

static HParsedToken*
act_LZW_clear(const HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	LZW_clear_table(ctx);
	//fprintf(debug, "clear code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fprintf(debug, "%lu ", p->ast->uint); // DEBUG
	//fflush(debug); // DEBUG
	ctx->next = 258; // Caution: moving this before the call to LZW_clear_table() will cause a memory leak
	return H_MAKE_BYTES(NULL, 0);
}

static bool
validate_LZW_9bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "9 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	//if (ctx->next < BITLIMIT_9) // DEBUG
	//	assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next < BITLIMIT_9);
}

static bool
validate_LZW_10bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "10 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	//if (ctx->next >= BITLIMIT_9 && ctx->next < BITLIMIT_10) // DEBUG
	//	assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next >= BITLIMIT_9 && ctx->next < BITLIMIT_10);
}

static bool
validate_LZW_11bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "11 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	//if (ctx->next >= BITLIMIT_10 && ctx->next < BITLIMIT_11) // DEBUG
	//	assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next >= BITLIMIT_10 && ctx->next < BITLIMIT_11);
}

static bool
validate_LZW_12bitcodeword(HParseResult *p, void *u)
{
	LZW_context_T * ctx = (LZW_context_T *) u;
	//fprintf(debug, "12 bit code: %lu, next: %u\n", p->ast->uint, ctx->next); // DEBUG
	//fflush(debug); // DEBUG
	//if (ctx->next >= BITLIMIT_11 && ctx->next < BITLIMIT_12) // DEBUG
	//	assert(H_CAST_UINT(p->ast) <= ctx->next); // DEBUG
	return (ctx->next >= BITLIMIT_11 && ctx->next < BITLIMIT_12);
}

static bool
validate_LZW_clear(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code == 256);
}

static bool
validate_LZW_eod(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code == 257);
}

static bool
validate_LZW_output(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	LZW_context_T *ctx = u;

	return (code != 256 && code != 257 && code < ctx->next);
}

static HParsedToken*
act_LZW_output(const HParseResult *p, void *u)
{
	HBytes * code_str;
	HBytes * last_str;
	uint64_t code = H_CAST_UINT(p->ast);
	uint8_t * output_token;
	uint8_t * next_entry_token;
	LZW_context_T * ctx = (LZW_context_T *) u;

	//fprintf(debug, "code: %lu, next: %u\n", code, ctx->next); // DEBUG
	//fprintf(debug, "%lu ", code); // DEBUG
	//fflush(debug); // DEBUG

	assert(ctx->next >= 258);
	assert(ctx->next <= 4096);
	assert(code < ctx->next);
	assert(code != 256);
	//assert(code != 257);
	// XXX due to the use of h_butnot in the grammar, we enter this
	// action on eod, only to discard the result. so just bail.
	if (code == 257)
		return NULL;

	/*
	 * Retrieve the output from the dictionary.
	 * This is what we'll wrap in a HBytes for returning.
	 */
	code_str = ctx->lzw_code_table[code];
	assert(code_str != NULL);
	assert(code_str->len > 0);

	/*
	 * Fill in the missing last character of a previously assigned code,
	 * if there is one.
	 */
	if (ctx->next > 258) {
		last_str = ctx->lzw_code_table[ctx->next - 1];
		((uint8_t *)last_str->token)[last_str->len - 1] =
		    code_str->token[0];
		    // XXX casting away the const. we know what we're doing.
		    // could avoid HBytes by using our own struct but come on.
		    // a different design might avoid byte arrays in the
		    // table altogether by storing just the last character
		    // and the code of the prefix.
	}

	/*
	 * Update the dictionary with a new entry that is missing the last
	 * character which we will only learn when we process the next code.
	 */
	next_entry_token = calloc(code_str->len + 1, sizeof(uint8_t));
	memcpy(next_entry_token, code_str->token, code_str->len);
	lzw_table_insert(ctx, next_entry_token, code_str->len + 1);

	/*
	 * Return a copy of the output.
	 */
	output_token = h_arena_malloc(p->arena, code_str->len);
	memcpy(output_token, code_str->token, code_str->len);
	return H_MAKE_BYTES(output_token, code_str->len);
}

static HParsedToken*
act_LZW_body(const HParseResult *p, void *u)
{
	size_t index = 0;
	size_t total_buffer_size = 0;
	size_t num_fragments = h_seq_len(p->ast);
	uint8_t * buffer;

	/* sum total bytes in array, alloc buffer */
	for(int i = 0; i < num_fragments; i++)
	{
		total_buffer_size += H_FIELD_BYTES(i).len;
	}

	buffer = h_arena_malloc(p->arena, sizeof(uint8_t) * total_buffer_size); // XXX arena alloc, calloc

	/* go through parse result, merge bytes */
	for(int i = 0; i < num_fragments; i++)
	{
		size_t len = H_FIELD_BYTES(i).len;
		memcpy(&buffer[index], H_FIELD_BYTES(i).token, len);
		index += len;
	}

	//fprintf(debug, "\n\n"); // DEBUG
	//fwrite(buffer, 1, total_buffer_size, debug); // DEBUG
	//fflush(debug); // DEBUG

	return H_MAKE_BYTES(buffer, total_buffer_size);
}


static HParsedToken*
act_LZW_data(const HParseResult *p, void *u)
{
	/* The AST this semantic action receives is a sequence that looks something like this:
		elements[0] -> TT_BYTES representing the initial clear code
		elements[1] -> TT_BYTES containing the decompressed data
		elements[2] -> TT_UINT representing the EOD code
	*/

	//HCountedArray * seq = H_CAST_SEQ(p->ast);
	//LZW_context_T *ctx = (LZW_context_T*) u; // DEBUG

	//fprintf(debug, "\n\n"); // DEBUG
	/*for(int i = 258; i < ctx->next; ++i) // DEBUG
	{
		fprintf(debug, "i: %u, str: ", i);
		fwrite(ctx->lzw_code_table[i]->token, ctx->lzw_code_table[i]->len, 1, debug);
		fprintf(debug, "\n");
	}
	fflush(debug); // DEBUG */

	return H_FIELD_TOKEN(1);
}


void init_LZW_parser()
{
	context = malloc(sizeof(LZW_context_T));
	memset(context, 0, sizeof(*context));
	context->next = 258;
	/* set up literals in LZW code table */
	for(int i = 0; i < 256; i++)
	{
		uint8_t *token = malloc(sizeof(uint8_t));
		*token = i;
		HBytes *lit = malloc(sizeof(HBytes));
		lit->token = token;
		lit->len = 1;
		context->lzw_code_table[i] = lit;
	}
	context->lzw_code_table[256] = calloc(1, sizeof(HBytes));
	context->earlychange = 1;

	H_VDRULE(LZW_9bitcodeword, h_bits(9, false), context);
	H_VDRULE(LZW_10bitcodeword, h_bits(10, false), context);
	H_VDRULE(LZW_11bitcodeword, h_bits(11, false), context);
	H_VDRULE(LZW_12bitcodeword, h_bits(12, false), context);
	H_RULE  (codeword,	h_choice(LZW_9bitcodeword, LZW_10bitcodeword,
				    LZW_11bitcodeword, LZW_12bitcodeword, NULL));

	H_AVDRULE(LZW_clear,	codeword, context);
	H_VDRULE (LZW_eod,	codeword, context);
	H_AVDRULE(LZW_output,	codeword, context);

	H_ADRULE(LZW_body, h_many(h_choice(LZW_clear, LZW_output, NULL)), context);

	H_ADRULE(LZW_data, h_sequence(LZW_clear, LZW_body, LZW_eod, NULL), context);
	    // XXX validate that the last byte is zero-padded?
	    // XXX require h_end_p()?

	p_lzwdata = LZW_data;
}


HParseResult* parse_LZW_data(const uint8_t* input, size_t length)
{
	//debug = fopen("lzw_debug.txt", "a"); // DEBUG
	HParseResult *res = h_parse(p_lzwdata, input, length);
	//fclose(debug); // DEBUG
	return res;
}


void init_LZW_context(int earlychange)
{
	for(int i = 258; i < 4096; ++i)
	{
		if(context->lzw_code_table[i] != NULL)
		{
			free((uint8_t *) context->lzw_code_table[i]->token); // These can be freed without issue, because HParsedTokens containing them have separate deep copies
			free(context->lzw_code_table[i]);
		}
		context->lzw_code_table[i] = NULL;
	}
	context->next = 258;
	context->earlychange = earlychange;
}
