/* beginnings of a PDF parser in hammer */

#include <string.h>	/* strncmp(), memset(), memcpy() */
#include <stdlib.h>	/* exit() */
#include <hammer/hammer.h>
#include <hammer/glue.h>
#include <math.h>

/* convenience macros */
#define SEQ(...)	h_sequence(__VA_ARGS__, NULL)
#define CHX(...)	h_choice(__VA_ARGS__, NULL)
#define OPT(X)		h_ignore(h_optional(X))
#define REP(P,N)	h_repeat_n(P, N)

#define IGN(P)		h_ignore(P)
#define LIT(S)		h_literal(S)
#define IN(STR)		h_in((const uint8_t *)(STR), sizeof(STR) - 1)
#define NOT_IN(STR)	h_not_in((const uint8_t *)(STR), sizeof(STR) - 1)

#ifdef LOG
#define VIOL(P,VIOL)	h_action(h_sequence(P, h_tell(), NULL), act_viol, VIOL)
#else
#define VIOL(P,VIOL)	P
#endif


/*
 * Works differently from strictness: messages with severity equal to or
 * greater than this number will be printed
 */
uint64_t log_level = 0;

/*
 * "Severity" is more for the sake of act_viol(), and I'm not sure it's a good
 * idea to use the same scale for other errors (such as error messages in
 * semantic actions other than act_viol). Use this to avoid having to assign
 * severity to an error message. Don't use it for violations.
 */
#define SEV_DONTCARE 99998

#include <stdarg.h>

void log_message(uint64_t severity, const char *format, ...)
{
	va_list ap;

	if(log_level <= severity) {
		va_start(ap, format);
		vfprintf(stderr, format, ap);
		va_end(ap);
	}
}


/*
 * some helpers
 */

HParser *p_fail;
HParser *p_epsilon;
HParser *p_return_0;
HParser *p_return_1;
uint8_t strictness = 255;	/* maximally lenient */

/* a combinator to parse a given character but return a different value */

HParsedToken *
act_return_uint(const HParseResult *p, void *u)
{
	return H_MAKE_UINT((uint64_t)(uintptr_t)u);
}

HParser *
p_mapch(uint8_t c, uint8_t v)
{
	return h_action(h_ch(c), act_return_uint, (void *)(uintptr_t)v);
}

/* a parser that just returns a given token */

HParsedToken *
act_return(const HParseResult *p, void *u)
{
	return u;
}

HParser *
p_return__m(HAllocator *mm__, const HParsedToken *tok)
{
	return h_action__m(mm__, p_epsilon, act_return, (void *)tok);
}

/* a parser that just returns a given uint */
HParser *
p_return_uint__m(HAllocator *mm__, uint64_t x)
{
	return h_action__m(mm__, p_epsilon, act_return_uint, (void *)x);
}

/* like h_sepBy but parses a fixed number of elements */
HParser *
p_sepBy_n__m(HAllocator *mm__, HParser *p, HParser *sep, size_t n)
{
	if (n == 0)
		return p_epsilon;

	HParser *sep_p = h_sequence__m(mm__, sep, p, NULL);
	HParser *tail = h_repeat_n__m(mm__, sep_p, n - 1);
	HParser *seq = h_sequence__m(mm__, p, tail, NULL);

	return h_action__m(mm__, seq, h_act_flatten, NULL);
}

/* a helper to compare an HBytes to a string */
bool
bytes_eq(HBytes b, const char *s)
{
	return strncmp(s, (const char *)(b.token), b.len) == 0 && b.len == strlen(s);
}

/* compare two HBytes like strcmp(3) */
int
bytescmp(HBytes b1, HBytes b2)
{
	size_t lmin = b1.len < b2.len ? b1.len : b2.len;
	int d;

	if ((d = memcmp(b1.token, b2.token, lmin)) != 0)
		return d;		/* strings distinguished by prefix */
	else if (b1.len < b2.len)
		return -1;		/* b1 is prefix of b2 ("b1 < b2") */
	else if (b1.len > b2.len)
		return 1;		/* b2 is prefix of b1 ("b1 > b2") */
	else
		return 0;		/* strings are equal */
}

bool
validate_eq_uint(HParseResult *p, void *u)
{
	const HParsedToken *v = p->ast;
	return (v != NULL && v->token_type == TT_UINT &&
	    v->uint == (uint64_t)(uintptr_t)u);
}

bool
validate_notnull(HParseResult *p, void *u)
{
	return p->ast != NULL;
}



// Forward declaration of Token structures
typedef struct { size_t nr, gen; } Ref;

typedef HCountedArray Dict;





// Forward declaration of text extraction related structures
struct textnode;
struct textstr;
struct TextEntry_S;
struct PtNode_S;

// ***********************************************************
/*
 * Text data structures
 */
typedef struct RsrcDict_S {
    const HParsedToken   *resources;            // font references dictionary (resources == NULL) ==> inherit
    const HParsedToken   *fonts;                // dictonary of fonts used in this page
    size_t                numFonts;
    const HParsedToken   *xobj;                 // xobj used in this page (?? is this <=1??, can page use multiple xobjects??)
//  Dict                **seenFonts;
//  size_t                numSeenFonts;
//  const HParsedToken  **seenCmaps;            // memoized cmaps (should this be a bytestream?
//  size_t                numCmapsSeen;
} RsrcDict_T;




struct textnode {
	struct textstr  *tstr;
	struct textnode *next;
};


struct fontref {
	const uint8_t *fontname;
	uint32_t       namelen;
	double         fontsize;
	char          *fn;   /* null terminated string */
};
struct textpos {
	double tx;
	double ty;
};

struct textmat { double cell[6]; };

struct textstr {
	uint8_t            *text;
	uint32_t            nchars;
	const HParsedToken *tobj;
};


struct tarrayelt {
	union {
		double          adj;
		struct textstr  tstr;
	};
	bool                isStr;
};
struct textarray {
	struct tarrayelt  *elts;
	uint32_t           nelts;
	struct textstr     flattened;
};

typedef struct {
    struct PtNode_S    *node;
    struct TextEntry_S *font;          // font name reference
    double              char_spacing;  // width adjustment for characters
    double              word_spacing;  // width adjustment for the space character
    double              line_spacing;  // line adjustment
    double              horiz_scaling; // line adjustment
    double              font_size;     // font size
    struct textpos      curr_pos;      // text position on page
} TextState_T;



/* operator::
 * TS -- Text state    : Table 105
 * TP -- Text position : Table 108
 * TW -- Test showing  : Table 109
 * */
typedef struct TextEntry_S {
	enum {TS_Tc, TS_Tw, TS_Tz, TS_TL, TS_Tf, TS_Tr, TS_Ts,
	      TP_Td, TP_TD, TP_Tm, TP_Tstar,
	      TW_Tj, TW_Tq, TW_Tqq, TW_TJ} type;
	//TextState_T	ts;	// text state associated with this string (TBD: other state attributes)
	struct PtNode_S	*node;	// page or XObject node the string is associated with
	union {
		double              value;	/* many just have a value */
		uint8_t             mode;	/* text mode */
		struct fontref      fref;	/* font name reference */
		struct textpos      pos;	/* text position */
		struct textmat      fm;		/* font matrix */
		struct {
			struct textstr tstr;	/* the string */
			double         aw;	/* opt. word spacing */
			double         ac;	/* opt. character spacing */
		};
		struct textarray    tarray;	/* text contained in an array object */
	};
	const HParsedToken *obj;
} TextEntry;  // text object entries


// Page Node
typedef struct PtNode_S {
    enum {PG_TREE, PG_NODE, XO_NODE} type;
    struct PtNode_S     *parent_n;              // reference to the parent node struct
    const HParsedToken  *parent_t;              // Type = Page tree -- reference
    const HParsedToken  *me_t;                  // Reference for me (*(Ref*)me->user ==> Ref structure)
    unsigned int         nodeNum;               // number in my parent's kids array
    RsrcDict_T          *pgRsrc;                // resource structure
    struct textpos       mediaBox;              // layout page dimensions
    TextState_T          ts;                    // text state associated with this node
    struct {
        double    width;                        // page width described by MediaBox
        double    height;                       // page height described by MediaBox
    } pagesz;
    union {
        struct {
            const Dict          *dict;          // page node dictionary
            unsigned int         page_num;      // number based on pgCount in the Catalog
            const HParsedToken  *textStream;    // content stream -- may be a result of concatenating array of content streams
        } pn;
        struct {
            struct PtNode_S     *kids;          // page tree array
            unsigned int         numPages;         // number of kids
       } pt;
        struct {
            const Dict          *dict;          // page node dictionary
            struct textpos       curr_pos;      // text position on page
            const HParsedToken  *textStream;    // content stream -- may be a result of concatenating array of content streams
        } xn;                                   // content within an xobject
    };

} PtNode_T;

struct XoNode_S;
typedef struct XoNode_S {
    char                *name;
    PtNode_T            *node;
    struct XoNode_S     *next;
} XoNode_T;


typedef struct Catalog_S {
    const  HParsedToken  *catalog;   // reference
    const  HParsedToken  *pRoot;     // reference
    PtNode_T              pgTree;    // page tree
    unsigned int          lastPage;  // incremented on visit of a page node
    unsigned int          lastNode;  // incremented on visit of any PtNode_T

    // TODO:: Determine if this should be moved to the PageTree
    // XoNode is one type of PageNode
    XoNode_T              xObjs;     // list of XObjects
    XoNode_T             *xoHead;
    XoNode_T             *xoTail;
    unsigned int          xoCount;   // number of xobjects -- not sure we use this
} Catalog_T;



// Initial use -- object streams
typedef struct {
	Ref                 oid;
	const HParsedToken  *obj;
} Objref_T;
typedef struct {
	size_t             numObjs;
	Objref_T           *tok;
	HArena             *arena;
} ObjStm;

typedef struct {
	char *name;
	char *type;
	char *basefont;
	char *encoding;
	const HParsedToken *descriptor;
	const HParsedToken *toUnicode;
	const HParsedToken *descendantFonts;
} Fontinfo_T;


// ***********************************************************


/*
 * auxiliary global data structure needed by the parser
 */
struct Env {
	const char *infile;
	const uint8_t *input;
	size_t sz;

	const HParsedToken **xrefs;	/* all xref sections of the file */
	size_t nxrefs;

	struct textnode     *txthead;  /* parsed text objects from the file */
	struct textnode     *txttail;  /* parsed text objects from the file */
	size_t               ntextobjs;

	Catalog_T            catalog;      /* Catalog object and document structure */
	PtNode_T             *curr_node;   /* current text state */

	unsigned int paren_nest_depth; /* String nesting depth */
	unsigned int array_nest_depth; /* Array nesting depth */
	unsigned int dict_nest_depth;  /* Dictionary nesting depth */

};

Fontinfo_T *lookup_font(TextState_T *state, struct Env *aux);

// ***********************************************************



/*
 * custom token types
 */
HTokenType TT_XREntry, TT_Ref, TT_Dict, TT_Null;
HTokenType TT_HParseResult;	// XXX what is this for?!
HTokenType TT_TextEntry, TT_ObjStm;

typedef struct {
	enum {XR_FREE, XR_INUSE, XR_OBJSTM} type;
	union {
		struct { size_t next, ngen; } f;	/* free */
		struct { size_t offs, gen; } n;		/* inuse */
		struct { size_t stm, idx; } o;		/* objstm */
	};
	const HParsedToken *obj;
} XREntry;


/* look up a value in a dictionary */
const HParsedToken *
dictentry(const Dict *dict, const char *key)
{
	HParsedToken *ent, *v;
	HBytes k;
	size_t len;

	if (dict == NULL)
		return NULL;

	len = strlen(key);
	for (size_t i = 0; i < dict->used; i++) {
		ent = dict->elements[i];
		k = H_INDEX_BYTES(ent, 0);

		if (h_seq_len(ent) < 2)	// XXX guarantee this in grammar
			continue;

		if (k.len == len && bytes_eq(k, key)) {
			v = H_INDEX_TOKEN(ent, 1);

			/* treat (direct) null entries as nonexistent */
			if (v->token_type == TT_Null)
				return NULL;
			else
				return v;
		}
	}

	return NULL;
}

void
pp_xrentry(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	XREntry *xr = H_CAST(XREntry, tok);

	switch (xr->type) {
	case XR_FREE:
		fprintf(stream, "{ \"type\":\"free\", \"next\":%zu, "
		    "\"ngen\":%zu }", xr->f.next, xr->f.ngen);
		break;
	case XR_INUSE:
		fprintf(stream, "{ \"type\":\"inuse\", \"offs\":%zu, "
		    "\"gen\":%zu }", xr->n.offs, xr->n.gen);
		break;
	case XR_OBJSTM:
		fprintf(stream, "{ \"type\":\"objstm\", \"stm\":%zu, "
		    "\"idx\":%zu }", xr->o.stm, xr->o.idx);
		break;
	default:
		assert(!"reached");
	}
}

void
pp_ref(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	Ref *r = H_CAST(Ref, tok);

	fprintf(stream, "[%zu,%zu]", r->nr, r->gen);
}

void
pp_parseresult(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	HParseResult *res = H_CAST(HParseResult, tok);

	if (res == NULL)
		fprintf(stream, "null");
	else
		h_pprint(stream, res->ast, indent, delta);
}

void
pp_dict(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	const HCountedArray *k_v;

	if (tok->seq->used == 0) {
		fprintf(stream, "{ }");
		return;
	}

	fprintf(stream, "{%*s", delta - 1, "");
	for (size_t i = 0; i < tok->seq->used; i++) {
		k_v = tok->seq->elements[i]->seq;
		if (i > 0)
			fprintf(stream, "\n%*s,%*s", indent, "", delta - 1, "");
		h_pprint(stream, k_v->elements[0], indent + delta, delta);
		fprintf(stream, ":");
		h_pprint(stream, k_v->elements[1], indent + delta, delta);
	}

	if (tok->seq->used > 2)
		fprintf(stream, "\n%*s}", indent, "");
	else
		fprintf(stream, " }\n");
}

void
pp_null(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	fprintf(stream, "null");
}

/*
 * semantic actions
 */

HParsedToken *
act_digit(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_CAST_UINT(p->ast) - '0');
}
#define act_pdigit act_digit
#define act_odigit act_digit

HParsedToken *
act_hlower(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(10 + H_CAST_UINT(p->ast) - 'a');
}

HParsedToken *
act_hupper(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(10 + H_CAST_UINT(p->ast) - 'A');
}

HParsedToken*
act_hdigitpair(const HParseResult *p, void *u)
{
	uint8_t b = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	size_t digits_processed = 0;
	uint8_t digits[2] = {0, 0};

	for(size_t i = 0; i < seq->used; ++i)
	{
		assert(seq->elements[i]->token_type == TT_UINT);
		digits[digits_processed] = H_CAST_UINT(seq->elements[i]);
		digits_processed++;
	}
	assert(digits_processed == 2);

	b = (digits[0] << 4) + digits[1];
	return H_MAKE_UINT(b);
}

HParsedToken *
act_ahextruncated(const HParseResult *p, void *u)
{
	uint8_t b = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	/* Assumption: At this point seq->elements[0] is a hex digit
	 * and seq->elements[1] holds '>' (EOD)
	 */
	assert(seq->used == 1);
	b = H_CAST_UINT(seq->elements[0]) << 4;
	return H_MAKE_UINT(b);
}

HParsedToken *
act_ahs_end(const HParseResult *p, void *u)
{
	HParsedToken *res;
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	assert(seq->used >= 1);

	res = H_MAKE_UINT(H_CAST_UINT(seq->elements[0]));
	return res;
}


HParsedToken *
act_ahexeod(const HParseResult *p, void *u)
{
	return NULL;
}


HParsedToken *
act_ahexstream(const HParseResult *p, void *u)
{
	uint8_t *result_bytes;
	size_t num_pairs;
	size_t required_bytes;
	HCountedArray *seq = H_CAST_SEQ(p->ast); // should be 2 -- digitpair, and ending
	HParsedToken *res;

	/* Ignore the the last element, which is EOD */
	/* Count up the number of bytes in the stream */
	assert(seq->used ==2 && seq->elements[0]->token_type == TT_SEQUENCE);
	HCountedArray *digitPairs = seq->elements[0]->seq;
	required_bytes = digitPairs->used;
	num_pairs      = required_bytes;
	HCountedArray *lastDigit = seq->elements[1]->seq;
	assert((lastDigit->used == 2) &&  // make sure that the last element is '>'
			(lastDigit->elements[1]->token_type == TT_UINT) &&
			(lastDigit->elements[1]->uint == 62));
	// account for truncated digit
	if (lastDigit->elements[0]->token_type == TT_UINT)
		required_bytes += 1;

	result_bytes = h_arena_malloc(p->arena, sizeof(uint8_t) * required_bytes);

	/* memcpy all but the last group's bytes into a single array */
	for (size_t i = 0; i < num_pairs; ++i)
	{
		result_bytes[i] = H_CAST_UINT(digitPairs->elements[i]);
	}
	if (num_pairs < required_bytes) {
		result_bytes[num_pairs] = H_CAST_UINT(lastDigit->elements[0]);
	}

	res = H_MAKE_BYTES(result_bytes, required_bytes);
	return res;
}




#include <limits.h>	/* UINT32_MAX */

HParsedToken *
act_a85zero(const HParseResult *p, void *u)
{
	static const uint8_t bs[] = {0, 0, 0, 0};
	return H_MAKE_BYTES(bs, 4);
}

HParsedToken *
act_a85digit(const HParseResult *p, void *u)
{
	uint8_t b = H_CAST_UINT(p->ast);
	b -= '!';

	/* At this point we have the base85 value of one byte out of 5 */
	return H_MAKE_UINT(b);
}

/* a NULL return from the action below signals an invalid encoding */
#define validate_a85five validate_notnull
#define validate_a85partial validate_notnull

HParsedToken *
act_a85partial(const HParseResult *p, void *u)
{
	/*
	 * the encoding process interprets 1 to 4 input bytes as a number in
	 * base 256. this number is then converted to base 85.
	 *
	 * the n input bytes are taken big-endian, filling the most significant
	 * digits and padded to 4 digits with k = 4 - n of trailing zeroes.
	 * thus the resulting number is a multiple of 256^k. call it f.
	 *
	 * this intermediate number is converted to n + 1 base-85 digits and
	 * its k least significant digits are cut off. note that these digits
	 * are not necessarily zero. consider this operation a rounding down to
	 * a multiple g of 85^k.
	 *
	 * to decode, the given base-85 digits are padded with k zeroes to form
	 * g and must be rounded back up to a multiple of 256^k to recover f.
	 * if the encoding was valid, this adds at most 85^k - 1 and cannot
	 * yield a number above 256^4 - 1 (= UINT32_MAX). also, the number 0
	 * must not be encoded as 5 bytes of '!'. if these conditions are met,
	 * the n most significant bytes of f are the original input.
	 */

	uint8_t *bytes;
	size_t l = h_seq_len(p->ast);		/* n + 1 */
	size_t n = l - 1;
	size_t k = 4 - n;
	size_t i;
	uint64_t M = 1 << (8 * k);		/* 256^k */
	uint64_t m = 1;				/* 85^k */
	uint64_t f, g = 0;

	/* convert the l-digit (big-endian) base-85 number to uint64 */
	for (i = 0; i < l; i++)
		g = g * 85 + H_FIELD_UINT(i);

	/* pad it to the full width of 5 base-85 digits */
	for (i = 0; i < k; i++)
		m = m * 85;
	g = g * m;

	/* round up to a multiple of M = 256^k to recover f */
	f = g + M - 1;
	f = f - f % M;

	/* validate that our f actually has a preimage under the encoding... */
	if (f - g >= m) {
		/* a valid encoding would not have been rounded down this far */
		log_message(9, "ASCII85Decode: invalid value\n");
		return NULL;
	}
	if (f > UINT32_MAX) {
		/* a valid encoding could not have been this high */
		log_message(9, "ASCII85Decode: value out of bounds\n");
		return NULL;
	}
	if (f == 0 && l == 5) {
		/* four null bytes must be encoded as "z", not "!!!!!" */
		log_message(9, "ASCII85Decode: invalid group \"!!!!!\"\n");
		return NULL;
	}

	/* drop the k least significant base-256 digits of f */
	f = f >> (8 * k);

	/* extract the n most significant digits as the original input */
	bytes = h_arena_malloc(p->arena, n);
	for (i = 0; i < n; i++) {
		bytes[n - 1 - i] = f;		/* mod 256 */
		f = f >> 8;			/* go to next digit */
	}
	assert(f == 0);

	return H_MAKE_BYTES(bytes, n);
}
#define act_a85five act_a85partial

HParsedToken *
act_ascii85(const HParseResult *p, void *u)
{
	HCountedArray *chunks = H_FIELD_SEQ(0);
	HParsedToken *partial = H_FIELD_TOKEN(1);
	HBytes chunk;
	uint8_t *buf;
	size_t n, i;

	n = chunks->used * 4;
	if (partial->token_type != TT_NONE)
		n += H_CAST_BYTES(partial).len;

	buf = h_arena_malloc(p->arena, n);

	/* copy all groups' bytes into a single array */
	for (i = 0; i < chunks->used; i++) {
		chunk = H_CAST_BYTES(chunks->elements[i]);
		assert(chunk.len == 4);
		memcpy(buf + i * 4, chunk.token, 4);
	}
	if (partial->token_type != TT_NONE) {
		chunk = H_CAST_BYTES(partial);
		assert(chunk.len < 4);
		memcpy(buf + chunks->used * 4, chunk.token, chunk.len);
	}

	return H_MAKE_BYTES(buf, n);
}

HParsedToken *
act_nat(const HParseResult *p, void *u)
{
	uint64_t x = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	for (size_t i = 0; i < seq->used; i++) {
		uint64_t d = H_CAST_UINT(seq->elements[i]);

		if (x > (UINT64_MAX - d) / 10)	/* would overflow */
			return NULL;
		// XXX introduce a structured error type and unify with VIOL()

		x = x * 10 + d;
	}

	return H_MAKE_UINT(x);
}
#define act_xroff act_nat
#define act_xrgen act_nat

#define validate_nat validate_notnull

HParser *p_violsev;
HParsedToken *
act_viol(const HParseResult *p, void *viol)
{
	uint32_t severity;
	uint32_t offset;
	HParseResult *severity_parse;
	viol = (uint8_t *) viol;
	severity_parse = h_parse(p_violsev, viol, strlen((char *)viol));
	if (!severity_parse) {
	        log_message(99999, "Severity for violation %s not assigned!\n", (char *)viol);
	        severity = 99999;
	}
	else {
	        severity = severity_parse->ast->seq->elements[0]->uint;
	}
	offset = p->ast->seq->elements[p->ast->seq->used-1]->uint / 8;
	log_message(severity, "VIOLATION[%d]@%d (0x%x): %s\n", severity, offset, offset, (char *) viol);
	if (severity > strictness) {
	        exit(1);
	}
	/* Just return the parse AST, drop the h_tell */
	if (p->ast->seq->used == 1) {
		return (HParsedToken *) NULL;
	}
	return (HParsedToken *) p->ast->seq->elements[0];
}

bool
validate_pnat(HParseResult *p, void *u)
{
	return H_CAST_UINT(p->ast) > 0;
}

HParsedToken *
act_sign(const HParseResult *p, void *u)
{
	return H_MAKE_SINT(p->ast == NULL ? 1 : -1);
}

/* p = (digits, period, digits) */
HParsedToken *
act_realnn(const HParseResult *p, void *u)
{
	double x = 0;
	double f = 0;
	HCountedArray *whole = H_FIELD_SEQ(0);
	HCountedArray *fract = H_FIELD_SEQ(2);

	// XXX check for overflow
	for (size_t i = 0; i < whole->used; i++)
		x = x*10 + H_CAST_UINT(whole->elements[i]);
	for (size_t i = 0; i < fract->used; i++)
		f = (f + H_CAST_UINT(fract->elements[fract->used - 1 - i])) / 10;

	return H_MAKE_DOUBLE(x + f);
}

bool
validate_intnn(HParseResult *p, void *u)
{
	return H_CAST_UINT(p->ast) <= INT64_MAX;
}

/* p = numbnn | (sign, numbnn) */
HParsedToken *
act_numb(const HParseResult *p, void *u)
{
	const HParsedToken *x = p->ast;
	int64_t sgn = 1;
	uint64_t abs_INT64_MIN = ((uint64_t) -(INT64_MIN + 1)) + 1;
		/* equals -INT64_MIN but avoids an overflow warning */

	if (x->token_type == TT_SEQUENCE) {
		sgn = H_FIELD_SINT(0);
		x = H_FIELD_TOKEN(1);
	}

	assert(sgn == 1 || sgn == -1);
	switch (x->token_type) {
	case TT_UINT:
		if ((sgn > 0 && x->uint > INT64_MAX) ||
		    (sgn < 0 && x->uint > abs_INT64_MIN))
			return NULL;	// XXX structured error type
		return H_MAKE_SINT(sgn * x->uint);
	case TT_DOUBLE:
		return H_MAKE_DOUBLE(sgn * x->dbl);
	default:
		assert(!"reached");
	}
}

#define validate_numb validate_notnull

HParsedToken *
act_null(const HParseResult *p, void *u)
{
	return H_MAKE(Null, NULL);
}

HParsedToken *
act_ref(const HParseResult *p, void *u)
{
	Ref *r = H_ALLOC(Ref);

	r->nr = H_FIELD_UINT(0, 0);
	r->gen = H_FIELD_UINT(0, 1);

	return H_MAKE(Ref, r);
}

HParsedToken *
act_bytes(const HParseResult *p, void *u)
{
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	uint8_t *bytes;

	bytes = h_arena_malloc(p->arena, seq->used);
	for (size_t i = 0; i < seq->used; i++)
		bytes[i] = H_CAST_UINT(seq->elements[i]);

	return H_MAKE_BYTES(bytes, seq->used);
}
#define act_token act_bytes
#define act_nstr act_bytes

HParsedToken *
act_nesc(const HParseResult *p, void *u)
{
	return H_MAKE_UINT(H_FIELD_UINT(1)*16 + H_FIELD_UINT(2));
}

#define act_schars h_act_flatten
#define act_string act_bytes

HParsedToken *
act_octal(const HParseResult *p, void *u)
{
	uint64_t x = 0;
	HCountedArray *seq = H_CAST_SEQ(p->ast);

	for (size_t i = 0; i < seq->used; i++)
		x = x*8 + H_CAST_UINT(seq->elements[i]);

	return H_MAKE_UINT(x);
}

HParsedToken *
act_xrent(const HParseResult *p, void *u)
{
	XREntry *xr = H_ALLOC(XREntry);
	char c = H_FIELD_UINT(2);

	xr->obj = NULL;
	switch (c) {
	case 'f':
		xr->type = XR_FREE;
		xr->f.next = H_FIELD_UINT(0);
		xr->f.ngen = H_FIELD_UINT(1);
		break;
	case 'n':
		xr->type = XR_INUSE;
		xr->n.offs = H_FIELD_UINT(0);
		xr->n.gen = H_FIELD_UINT(1);
		break;
	default:
		assert(!"reached");
	}

	return H_MAKE(XREntry, xr);
}

HParsedToken *
act_xrstment(const HParseResult *p, void *u)
{
	XREntry *xr = H_ALLOC(XREntry);

	xr->obj = NULL;
	xr->type = H_FIELD_UINT(0);
	switch (xr->type) {
	case XR_FREE:
		xr->f.next = H_FIELD_UINT(1);
		xr->f.ngen = H_FIELD_UINT(2);
		break;
	case XR_INUSE:
		xr->n.offs = H_FIELD_UINT(1);
		xr->n.gen = H_FIELD_UINT(2);
		break;
	case XR_OBJSTM:
		xr->o.stm = H_FIELD_UINT(1);
		xr->o.idx = H_FIELD_UINT(2);
		break;
	default:
		log_message(0, "unknown entry type %d in xref stream",
		    xr->type);
		/* NB: this is allowed and treated as null by resolve() */
	}

	return H_MAKE(XREntry, xr);
}

#define act_xrefs h_act_last

/*
 * return a cross-reference stream in the same form as xref sections.
 *
 * p = ((pnat nat) (dict [HParseResult: xrefs]))
 * result = (xrefs dict)
 */
HParsedToken *
act_xrstm(const HParseResult *p, void *u)
{
	const HParsedToken *xrefs, *dict;
	HParsedToken *tok;
	HParseResult *res;

	dict = H_INDEX_TOKEN(p->ast, 1, 0);
	res = H_FIELD(HParseResult, 1, 1);	// XXX free this
	xrefs = res != NULL ? res->ast : NULL;

	tok = H_MAKE_SEQN(2);
	tok->seq->elements[0] = (HParsedToken *)xrefs;
	tok->seq->elements[1] = (HParsedToken *)dict;
	tok->seq->used = 2;
	return tok;
}

/*
 * validate the presence of an AST holding xref entries,
 * (i. e. a successful parse of an xref table),
 * and the /Type field on a cross-reference stream
 *
 * p = ((pnat nat) (dict offs offs))
 */
bool
validate_xrstm(HParseResult *p, void *u)
{
	const HParsedToken *xrefs = H_INDEX_TOKEN(p->ast, 1, 1);
	const Dict *tdict = H_FIELD(Dict, 1, 0);
	const HParsedToken *v = dictentry(tdict, "Type");

	if (!xrefs)
	{
		return false;
	}

#if 0
	if (v == NULL)
		fprintf(stderr, "stream dict has no /Type\n");
	else if (v->token_type != TT_BYTES)
		fprintf(stderr, "stream /Type is no name object\n");
	else if (bytes_eq(v->bytes, "XRef"))
		return true;
	return false;
#endif	// XXX this block can be removed

	return (v != NULL && v->token_type == TT_BYTES &&
	    bytes_eq(v->bytes, "XRef"));
}

HParsedToken *
act_dict_(const HParseResult *p, void *env)
{
	Dict *dict = H_CAST_SEQ(p->ast);

	/*
	 * NB: Null values in dictionaries are to be treated as nonexistent
	 * entries (see ISO 32000-2 7.3.9 "Null object"). However, to avoid
	 * complicating our parser, we leave it to dictentry() to drop them.
	 *
	 * Note further that null values can occur indirectly, so resolve()
	 * also maps them to NULL, equivalent to a nonexistent object.
	 */

	return H_MAKE(Dict, dict);
}

bool
validate_dict_(HParseResult *p, void *u)
{
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	HBytes ki, kj;
	size_t i, j;

	/* validate that there are no duplicate keys */
	// XXX make this more efficient than quadratic runtime
	for (i = 0; i < seq->used; i++) {
		ki = H_INDEX_BYTES(seq->elements[i], 0);

		for (j = 0; j < i; j++) {
			kj = H_INDEX_BYTES(seq->elements[j], 0);

			if (bytescmp(ki, kj) == 0) {
				log_message(7, "duplicate key /%*s in "
				    "dictionary\n", (int)ki.len, ki.token);
				return false;
			}
		}
	}

	return true;
}

#define act_array_ h_act_flatten

HParsedToken *
act_shortlength(const HParseResult *p, void *u)
{
	uint8_t length = H_CAST_UINT(p->ast);
	/* Length can range from 0-127, corresponding to the range 1-128, inclusive */
	uint8_t finallength = length+1;

	return H_MAKE_UINT(finallength);
}

HParsedToken *
act_longlength(const HParseResult *p, void *u)
{
	uint8_t length = H_CAST_UINT(p->ast);
	uint8_t finallength = 257-length;

	return H_MAKE_UINT(finallength);
}

HParsedToken *
act_longrun(const HParseResult *p, void *u)
{
	HParsedToken **elements = h_seq_elements(p->ast);
	HParsedToken *res = H_MAKE_SEQ();

	uint8_t length = H_CAST_UINT(elements[0]);
	uint8_t data = H_CAST_UINT(elements[1]);

	for (size_t len = 0; len < length; ++len)
	{
		h_seq_snoc(res, H_MAKE_UINT(data));
	}

	return res;
}

HParsedToken *
act_rldstring(const HParseResult *p, void *u)
{
	const HParsedToken *flattened = h_seq_flatten(p->arena, p->ast);
	HCountedArray *flattened_seq = H_CAST_SEQ(flattened);
	size_t bytes_required;
	uint8_t *result_bytes;

	if (flattened_seq->used)
	{
		bytes_required = flattened_seq->used - 1;
		result_bytes = h_arena_malloc(p->arena, sizeof(uint8_t) * bytes_required);

		for (size_t i = 0; i < flattened_seq->used-1; ++i)
		{
			result_bytes[i] = H_CAST_UINT(flattened_seq->elements[i]);
		}

		return H_MAKE_BYTES(result_bytes, bytes_required);
	}
	else
	{
		return H_MAKE_BYTES(NULL, 0);
	}
}

/*
 * The following functions implement a fix for being able to exhaust memory by nesting strings, objects and dicts too deeply.
 * A counter in aux (separately for each type of nesting) is incremented on each opening token, and decremented on each closing token. 
 * The counters are not decremented when they are 0 to avoid underflow.
 *
 * The counters are global across the document (but respect document structure,
 * e.g. parentheses in streams don't count, unless for some reason they're parsed with the lparen rule.)
 * The sopen, dopen, and array_begin validations make the parse fail if this nesting depth is exceeded.
 * Because currently there are no diagnostic messages, this can probably result in unexpected parses.
 *
 * validate_rparen, validate_array_end, and validate_dclose will make the parse fail if a respective closing token ( ')', ']', '>>' )
 * is encountered before an sopen, array_begin, or dopen (or there are more closing parentheses than opening ones.) This is not exactly
 * the same as balanced parentheses: the order different types of parentheses appear in is not considered. Different brackets can be considered
 * independently because other H_RULEs already reject cases with interleaved opening/closing tokens such as "<< [ /Foo /Bar >> ]".
 */

#define PAREN_MAX_NEST_DEPTH 256

bool
validate_sopen(HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	return aux->paren_nest_depth < PAREN_MAX_NEST_DEPTH;
}

HParsedToken *
act_sopen(const HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	aux->paren_nest_depth += 1;

	return H_MAKE_UINT(H_CAST_UINT(p->ast));
}

bool
validate_sclose(HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	return aux->paren_nest_depth > 0;
}

HParsedToken*
act_sclose(const HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	if(aux->paren_nest_depth > 0)
		aux->paren_nest_depth -= 1;

	return H_MAKE_UINT(H_CAST_UINT(p->ast));
}

#define ARRAY_MAX_NEST_DEPTH 256

bool
validate_array_begin(HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	return aux->array_nest_depth < ARRAY_MAX_NEST_DEPTH;
}

HParsedToken *
act_array_begin(const HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	aux->array_nest_depth += 1;

	return H_MAKE_UINT(H_CAST_UINT(p->ast));
}

bool
validate_array_end(HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	return aux->array_nest_depth > 0;
}

HParsedToken *
act_array_end(const HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	if(aux->array_nest_depth > 0)
		aux->array_nest_depth -= 1;

	return H_MAKE_UINT(H_CAST_UINT(p->ast));
}

#define DICT_MAX_NEST_DEPTH 256

bool
validate_dopen(HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	return aux->dict_nest_depth < DICT_MAX_NEST_DEPTH;
}

HParsedToken*
act_dopen(const HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	aux->dict_nest_depth += 1;

	return H_MAKE_BYTES(H_CAST_BYTES(p->ast).token, H_CAST_BYTES(p->ast).len);
}

bool
validate_dclose(HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	return aux->dict_nest_depth > 0;
}

HParsedToken*
act_dclose(const HParseResult *p, void *u)
{
	struct Env *aux = (struct Env*) u;
	if(aux->dict_nest_depth > 0)
		aux->dict_nest_depth -= 1;

	return H_MAKE_BYTES(H_CAST_BYTES(p->ast).token, H_CAST_BYTES(p->ast).len);
}

/*
 * ********************************************************************
 * Start Text parsing
 * ********************************************************************
 */
void pp_fontstate(FILE *stream, const TextState_T *state)
{
	assert(state);
	fprintf(stream, "\nFont State: Page = ");
	if ( (state->node) &&
			( (state->node->type == PG_NODE) || (state->node->type == XO_NODE) ) )
		pp_ref(stream, state->node->me_t, 0, 0);
	else
		fprintf(stream, ", Text not associated with a page or xobject\n");
	if (state->font)  // not all operators need or set this
		fprintf(stream, ", Font = %s\n", state->font->fref.fn);
	else
		fprintf(stream, ", Font not yet specified\n");

}
void pp_fontinfo(FILE *stream, const TextState_T *state, const Fontinfo_T *fi)
{
	assert(state && fi);
	pp_fontstate(stream, state);
	fprintf(stream, "Font Info: ");
	if (fi->name)     fprintf(stream, "Font= %s", fi->name);
	if (fi->type)     fprintf(stream, ", Type= %s", fi->type);
	if (fi->basefont) fprintf(stream, ", Basefont= %s", fi->basefont);
	if (fi->encoding)     fprintf(stream, ", Encoding= %s", fi->encoding);
	if (fi->descriptor) pp_dict(stream, fi->descriptor, 0, 0);
	if (fi->toUnicode) pp_dict(stream, fi->toUnicode, 0, 0);
	if (fi->descendantFonts) pp_dict(stream, fi->descendantFonts, 0, 0);
	fprintf(stream, "\n");
}

/*
 * Pretty printer for text components of the ast
 */
void
pp_textentry(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	TextEntry *txte = H_CAST(TextEntry, tok);

	switch (txte->type) {
	/*
	 * Always pretty print the text show operators
	 *
	 * If TEXT_VERBOSE is set, pretty-print the other operators
	 */
#define TEXT_VERBOSE
#ifdef TEXT_VERBOSE
	case TS_Tf:
		fprintf(stream, "Tf_op: fn=%.*s, fontsize=%3.3f\n",
				txte->fref.namelen, txte->fref.fontname, txte->fref.fontsize);
		break;
	case TP_Td:
		fprintf(stream, "Td_op: text position ::tx=%3.3f:ty=%3.3f\n",
				txte->pos.tx, txte->pos.ty);
		break;
#endif
	case TW_Tj:
	case TW_Tq:
	case TW_Tqq:
		fprintf(stream, "len=%u, ", txte->tstr.nchars);
		fwrite((const void *)txte->tstr.text, (int) txte->tstr.nchars, 1, stream);
		break;
	case TW_TJ:
		fprintf(stream, "len=%u, ", txte->tarray.flattened.nchars);
		fwrite((const void *)txte->tarray.flattened.text,
				(int) txte->tarray.flattened.nchars, 1, stream);
		break;


	default:
		;
	}
}


void
pp_objstm(FILE *stream, const HParsedToken *tok, int indent, int delta)
{
	ObjStm *entry = H_CAST(ObjStm, tok);

//	pp_parseresult(stream, (const HParsedToken *)entry->res, 0, 0);
	fprintf(stream, "pp_objstm:: Num Objects = %lu\n", entry->numObjs);

	for (int i=0; i<entry->numObjs; i++) {
		fprintf(stream, "oid: <nr=%lu, gen=%lu>, ", entry->tok[i].oid.nr, entry->tok[i].oid.gen);
		fprintf(stream, "Type = %u, ", entry->tok[i].obj->token_type);
		if (entry->tok[i].obj->token_type == TT_Dict) {
			pp_dict(stream, entry->tok[i].obj, 0, 0);
		}
		fprintf(stream, "\n");
	}
}

/*
 * semantic actions
 */



/*
 *  Simplify the code by casting the choice of integer number and real number to double
 */
bool
validate_tnumb(HParseResult *p, void *u)
{
	assert((p->ast->token_type == TT_SINT) || (p->ast->token_type == TT_DOUBLE));

	return true;
}


HParsedToken *
act_tnumb(const HParseResult *p, void *u)
{

	double value;

	if (p->ast->token_type == TT_SINT)
		value = p->ast->sint;
	else
		value = H_CAST_DOUBLE(p->ast);

	return H_MAKE_DOUBLE(value);
}




/*
 * Text state operators - Table 105
 *   TS_Tc, TS_Tw, TS_Tz, TS_TL, TS_Tf, TS_Tr, TS_Ts
 *
 * *****************************************************************
 * *****************************************************************
 *
 */
HParsedToken *
act_Tc_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;
	const HParsedToken *tval = H_INDEX_TOKEN(p->ast, 0);


	txte->type  = TS_Tc;
	txte->obj   = NULL;

	assert(tval->token_type == TT_DOUBLE);
	txte->value   = tval->dbl;

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 * Tw operator: word spacing specification
 * H_ARULE(Tw_op, SEQ(tnumb, ws, LIT("Tw")));  // 9.3.3 - wordSpace
 */
HParsedToken *
act_Tw_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type  = TS_Tw;
	txte->obj   = NULL;
	txte->value = H_FIELD_DOUBLE(0);

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}



/*
 * Tz operator: horizontal scaling specification
 * H_ARULE(Tz_op, SEQ(tnumb, ws, LIT("Tz")));  // 9.3.4 - horizontal scaling
 */
HParsedToken *
act_Tz_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type  = TS_Tz;
	txte->obj   = NULL;
	txte->value = H_FIELD_DOUBLE(0);

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}



/*
 * TL operator: leading (line spacing) specification
 * H_ARULE(TL_op, SEQ(tnumb, ws, LIT("TL")));  // 9.3.5 - leading
 */
HParsedToken *
act_TL_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type  = TS_TL;
	txte->obj   = NULL;
	txte->value = H_FIELD_DOUBLE(0);

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 *  Font name and size specification
 *  H_ARULE(Tf_op, SEQ(name, ws, nat, ws, KW("Tf"), ws));  // font and size
 *
 *  TODO: Verify that the name is specified in the resource dictionary
 */
HParsedToken *
act_Tf_op(const HParseResult *p, void *u)
{
	TextEntry          *txte     = H_ALLOC(TextEntry);
	struct Env         *aux      = (struct Env*)u;
	const HParsedToken *fn_token = H_FIELD_TOKEN(0);

	txte->type  = TS_Tf;
	txte->obj   = NULL;

	txte->fref.fontname = fn_token->bytes.token;
	txte->fref.namelen  = fn_token->bytes.len;
	txte->fref.fn       = h_arena_malloc(p->arena, sizeof (char) * (fn_token->bytes.len + 1));
	memcpy(txte->fref.fn, fn_token->bytes.token, fn_token->bytes.len);
	txte->fref.fn[fn_token->bytes.len] = '\0';
	HTokenType tokenType = p->ast->seq->elements[1]->token_type;
	if (tokenType == TT_UINT)
		txte->fref.fontsize = (double) H_FIELD_UINT(1);
	else if (tokenType == TT_DOUBLE)
		txte->fref.fontsize = (double) H_FIELD_DOUBLE(1);

    // associate the text with the current state
    txte->node = aux->curr_node;


	return H_MAKE(TextEntry, txte);
}



/*
 * Tr operator: rendering mode
 * H_VRULE(tmode, nat);                        // True if <= 7
 * H_ARULE(Tr_op, SEQ(tmode, ws, LIT("Tr")));  // 9.3.6 - rendering mode
 *
 *
 */
#define TEXTMODE_MAX 7

bool
validate_tmode(HParseResult *p, void *u)
{
	return H_CAST_UINT(p->ast) <= TEXTMODE_MAX;
}

HParsedToken *
act_Tr_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type  = TS_Tr;
	txte->obj   = NULL;
	txte->mode  = H_FIELD_UINT(0);

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 * Ts operator: rise specification
 * H_ARULE(Ts_op, SEQ(tnumb, ws, LIT("Ts")));  // rise
 */
HParsedToken *
act_Ts_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type  = TS_Ts;
	txte->obj   = NULL;
	txte->value = H_FIELD_DOUBLE(0);

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}



/*
 * 9.4.2 - Text positioning operators - Table 108
 *   TP_Td, TP_TD, TP_Tm, TP_Tstar
 *
 * *****************************************************************
 * *****************************************************************
 *
 * TP_Td: String position - Translation specification
 * H_ARULE(Td_op, SEQ(tnumb, ws, tnumb, ws, LIT("Td"), ws));   // move to next line with offset
 */
HParsedToken *
act_Td_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type   = TP_Td;
	txte->obj    = NULL;
	txte->pos.tx = H_FIELD_DOUBLE(0);
	txte->pos.ty = H_FIELD_DOUBLE(1);

	// associate the text with the current state
	// NOTE: This operator does not require a font
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}




/*
 * TP_TD: Offset to next line and set the leading parameter state
 * H_ARULE(TD_op, SEQ(tnumb, ws, tnumb, ws, LIT("TD")));               // move to next line with offset and set state
 */
HParsedToken *
act_TD_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;


	txte->type   = TP_TD;
	txte->obj    = NULL;
	txte->pos.tx = H_FIELD_DOUBLE(0);
	txte->pos.ty = H_FIELD_DOUBLE(1);

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 * TP_Tm: Text matrix specification
 * H_ARULE(Tm_op, SEQ(REP(SEQ(tnumb, ws), 6), LIT("Tm"), ws));    // set text matrix
 */
HParsedToken *
act_Tm_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
    struct Env         *aux    = (struct Env*)u;


	txte->type  = TP_Tm;
	txte->obj   = NULL;

	assert((p->ast->token_type == TT_SEQUENCE) &&
			(p->ast->seq->elements[0]->token_type == TT_SEQUENCE) &&
			(p->ast->seq->elements[0]->seq->used == 6));
	for (int i=0; i<6; i++)

		txte->fm.cell[i] = p->ast->seq->elements[0]->seq->elements[i]->seq->elements[0]->dbl;

    // associate the text with the current state
    txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 * TP_Tstar: Move to the next line
 * H_ARULE(Tstar_op, SEQ(LIT("T*"), ws));                     // move to next line
 */
HParsedToken *
act_Tstar_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	struct Env         *aux    = (struct Env*)u;

	txte->type  = TP_Tstar;
	txte->obj   = NULL;
	txte->value = 0;

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}



/*
 * 9.4.3 - Text showing operators - Table 109
 *   TW_Tj, TW_Tq, TW_Tqq, TW_TJ
 *
 * *****************************************************************
 * *****************************************************************
 *
 * TW_Tj: Show string
 * H_ARULE(Tj_op, SEQ(string, ws, LIT("Tj"), ws));          // show text string
 */
HParsedToken *
act_Tj_op(const HParseResult *p, void *u)
{
	TextEntry          *txte   = H_ALLOC(TextEntry);
	const HParsedToken *tstr   = H_INDEX_TOKEN(p->ast, 0);
	struct Env         *aux    = (struct Env*)u;


	txte->type  = TW_Tj;
	txte->obj   = NULL;

	txte->tstr.text   = (uint8_t *)tstr->bytes.token;
	txte->tstr.nchars = tstr->bytes.len;

	// associate the text with the current state
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 * TW_Tq: Offset to next line then show string
 * H_ARULE(TsingleQ_op, SEQ(string, ws, LIT(quote), ws));   // Move to next line and show text
 */
HParsedToken *
act_TsingleQ_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	const HParsedToken *tstr = H_INDEX_TOKEN(p->ast, 0);
	struct Env          *aux = (struct Env*)u;


	txte->type  = TW_Tq;
	txte->obj   = NULL;

	txte->tstr.text   = (uint8_t *)tstr->bytes.token;
	txte->tstr.nchars = tstr->bytes.len;

	// associate the text wth the current font
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}


/*
 * TW_Tqq: Offset to next line then show string, apply formatting specifications
 * H_ARULE(TdoubleQ_op, SEQ(tnumb, ws, tnumb, ws, string, ws, LIT(dquote), ws)); // Move to next line and show formatted text
 *
 */
HParsedToken *
act_TdoubleQ_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	const HParsedToken *aw = H_INDEX_TOKEN(p->ast, 0);
	const HParsedToken *ac = H_INDEX_TOKEN(p->ast, 1);
	const HParsedToken *tstr = H_INDEX_TOKEN(p->ast, 2);
	struct Env         *aux  = (struct Env*)u;


	txte->type  = TW_Tqq;
	txte->obj   = NULL;

	txte->aw          = aw->dbl;
	txte->ac          = ac->dbl;
	txte->tstr.text   = (uint8_t *)tstr->bytes.token;
	txte->tstr.nchars = tstr->bytes.len;

	// associate the text wth the current font
	txte->node = aux->curr_node;

	return H_MAKE(TextEntry, txte);
}

/*
 * TW_TJ: Show array of strings, with potentially re-positioning specifications for each string
 * H_RULE(TArr_elem, SEQ(OPT(SEQ(nanumbs)), string, ws))
 * H_ARULE(TJ_op, SEQ(h_many(TArr_elem), LIT("TJ"), ws));  // show one or more text strings
 *
 * TODO:: Implement the array parser
 */
HParsedToken *
act_TJ_op(const HParseResult *p, void *u)
{
	TextEntry          *txte = H_ALLOC(TextEntry);
	const HParsedToken *tarr = H_INDEX_TOKEN(p->ast, 0);
	struct Env         *aux  = (struct Env*)u;


	txte->type  = TW_TJ;
	txte->obj   = NULL;


	// associate the text wth the current font
	txte->node = aux->curr_node;

	/*
	 * Parse each element of the array
	 * Build up the pointers to each of the string pieces
	 */
	txte->tarray.nelts = tarr->seq->used;
	txte->tarray.elts  = h_arena_malloc(p->arena, sizeof(struct tarrayelt) * txte->tarray.nelts);
	txte->tarray.flattened.nchars = 0;

	for (int i=0; i<txte->tarray.nelts; i++) {
		const HParsedToken *elt = tarr->seq->elements[i];
		assert( (elt->token_type == TT_SEQUENCE) && (elt->seq->used == 1) );
		switch (elt->seq->elements[0]->token_type) {
		case TT_DOUBLE:
			txte->tarray.elts[i].adj         = elt->seq->elements[0]->dbl;
			txte->tarray.elts[i].isStr       = false;
			break;
		case TT_BYTES:
			txte->tarray.elts[i].tstr.text   = (uint8_t *)elt->seq->elements[0]->bytes.token;
			txte->tarray.elts[i].tstr.nchars = elt->seq->elements[0]->bytes.len;
			txte->tarray.elts[i].isStr       = true;
			txte->tarray.flattened.nchars   += txte->tarray.elts[i].tstr.nchars;

			break;
		default:
			log_message(SEV_DONTCARE, "act_TJ_op:: Unexpected element type :: %d\n", elt->seq->elements[0]->token_type);
			exit(2);
		}
	}

	/* hold on to a flattened copy of the string */
	txte->tarray.flattened.text = h_arena_malloc(p->arena, sizeof(char) * txte->tarray.flattened.nchars);
	int j = 0; // current index
	for (int i=0; i<txte->tarray.nelts; i++) {
		if (txte->tarray.elts[i].isStr) {
			memcpy(&txte->tarray.flattened.text[j], txte->tarray.elts[i].tstr.text, txte->tarray.elts[i].tstr.nchars);
			j += txte->tarray.elts[i].tstr.nchars;
		}
	}

	return H_MAKE(TextEntry, txte);
}




/*
 * Parse the text object delimited by "BT" and "ET"
 * Text Matrix and Text Line Matrix are reinitialized at each invocation
 *
 * TODO:: Maintain line number to vertically (horizontally) sort the text
 */
HParsedToken *
act_txtobj(const HParseResult *p, void *u)
{
	// we will need to handle the case when the text string is longer than 8192 characters
	struct Env         *aux    = (struct Env*)u;
	TextEntry          *txtobj = H_ALLOC(TextEntry);
	TextEntry          *txte   = NULL;
	const HParsedToken *opstream = H_INDEX_TOKEN(p->ast, 1);
	const HParsedToken *tt_text=NULL;
	uint8_t            *tstr=NULL;
	int                 textlen=0;
	struct textmat      tm;
	PtNode_T           *node = aux->curr_node;
	double              cs = node->ts.char_spacing;
	double              ws = node->ts.word_spacing;
	double              ls = node->ts.line_spacing;
	double             *px=&tm.cell[4];
	double             *py=&tm.cell[5];

	// initialize the text matrix
	tm.cell[0] = 1;
	tm.cell[1] = 0;
	tm.cell[2] = 0;
	tm.cell[3] = 1;
	tm.cell[4] = 0;
	tm.cell[5] = 0;

	// if the current page/XObject is not set, return a NULL token
	if (aux->curr_node == NULL) {
	return ((HParsedToken *) NULL);
	}

	// TODO:: Handle non-horizontal text


	// Walk through the tokens to determine how much space to allocate
	// Count the number of characters in the stream
	// Concatenate the text into the allocated space
	for (int i =0; i < opstream->seq->used; i++) {

		txte = H_CAST(TextEntry, opstream->seq->elements[i]);

		// make sure we are working on the same node as the current node
		assert(txte && (txte->node == node));

		switch (txte->type) {
		// text state operators
		case TS_Tc:
			node->ts.char_spacing = txte->value;
			break;

		case TS_Tw:
			node->ts.word_spacing = txte->value;
			break;

		case TS_Tz:
			node->ts.horiz_scaling = txte->value;
			break;

		case TS_TL:
			node->ts.line_spacing = txte->value;
			break;

		case TS_Tf:
			node->ts.font = txte;
			node->ts.font_size = txte->fref.fontsize;
			break;


		// text positioning and showing operators
		case TP_TD:
			node->ts.line_spacing = txte->pos.ty;
		case TP_Td:
			if ( (*px == 0.0) && (*py == 0.0) ) { // initialize
				*px = txte->pos.tx;
				*py = txte->pos.ty;
				// check to see if we are starting a new line
				if ( (node->ts.curr_pos.ty != 0.0) &&
				    (node->ts.curr_pos.ty != *py) ) {
					textlen += 1; // add a newline
				}
			} else {
				if (txte->pos.ty != 0.0) {
					// we are not rendering -- we just know
					// it is not in the same line if y not
					// equal
					textlen += 1; // add a newline
					*py -= txte->pos.ty; // should this be a +=??
				}
				if (txte->pos.tx) { // handle x -- when should we add a space
					// TODO:: handle x -- not sure .. for now, ignore
					*px += txte->pos.tx;
				}
			}
			break;
		case TP_Tstar:
			*py -= node->ts.line_spacing;
			textlen += 1;
			break;


		case TW_Tqq:
			node->ts.word_spacing = txte->aw;
			node->ts.char_spacing = txte->ac;
		case TW_Tq:
			*py -= node->ts.line_spacing;
			textlen += 1;
		case TW_Tj:
			textlen += txte->tstr.nchars;
			if (node->ts.font != NULL)
				*px += txte->tstr.nchars * node->ts.font->fref.fontsize; // TODO:: handle character width from font description
			break;

		case TW_TJ:
			textlen += txte->tarray.flattened.nchars;
			if (node->ts.font != NULL)
				*px += txte->tarray.flattened.nchars * node->ts.font->fref.fontsize; // TODO:: handle character width from font description
			break;

		default:
			; // ignore
		}
	}

	// Are we within the page bounds? If not, generate a warning
	if ( (*px < 0) || (*px > aux->curr_node->mediaBox.tx) ) {
		fprintf (stdout, "Final position of the text string is outside media box bounds.\n"
		    "Media Box-page width=%f, String end position-@width=%f\n",
		    aux->curr_node->mediaBox.tx, *px);
	}
	if ( (*py < 0) || (*py > aux->curr_node->mediaBox.ty) ) {
		fprintf (stdout, "Final position of the text string is outside media box bounds.\n"
		    "Media Box-page height=%f, String end position-@height=%f\n",
		    aux->curr_node->mediaBox.tx, *px);
	}

	// reset text state
	*px = *py = 0.0;
	node->ts.char_spacing = cs;
	node->ts.word_spacing = ws;
	node->ts.line_spacing = ls;

	tstr = h_arena_malloc(p->arena, sizeof(uint8_t) * textlen);
	int idx=0;
	TextState_T      *ts;
	// Now concatenate the pieces
	for (int i =0; i < opstream->seq->used; i++) {
		txte = H_CAST(TextEntry, opstream->seq->elements[i]);
		ts   = &txte->node->ts;

		// Process the text operators
		switch (txte->type) {
		// text state operators
		case TS_Tc:
			node->ts.char_spacing = txte->value;
			break;

		case TS_Tw:
			node->ts.word_spacing = txte->value;
			break;

		case TS_Tz:
			node->ts.horiz_scaling = txte->value;
			break;

		case TS_TL:
			node->ts.line_spacing = txte->value;
			break;

		case TS_Tf:
			ts->font      = txte;
			node->ts.font_size = txte->fref.fontsize;
			break;


		// text positioning and showing operators
		case TP_TD:
		case TP_Td:
			if ( (*px == 0.0) && (*py == 0.0) ) { // initialize
				*px = txte->pos.tx;
				*py = txte->pos.ty;
				if ( (node->ts.curr_pos.ty != 0.0) &&
				    (node->ts.curr_pos.ty != *py) ) {
					tstr[idx] = '\n'; idx += 1;
				}
			} else {
				if (txte->pos.ty != 0.0) {
					tstr[idx] = '\n'; idx += 1;
					*py -= txte->pos.ty; // should this be a +=??
				}
				if (txte->pos.tx) { // handle x -- when should we add a space
					// TODO:: handle x -- not sure .. for now, ignore
					*px += txte->pos.tx;
				}
			}
			break;

		case TP_Tstar:
			tstr[idx] = '\n'; idx += 1;
			*py -= node->ts.line_spacing;
			break;


		case TW_Tqq:
			node->ts.word_spacing = txte->aw;
			node->ts.char_spacing = txte->ac;
		case TW_Tq:
			*py -= node->ts.line_spacing;
			tstr[idx] = '\n'; idx += 1;
		case TW_Tj:
			memcpy(&tstr[idx], txte->tstr.text, txte->tstr.nchars);
			idx += txte->tstr.nchars;
			if (node->ts.font != NULL)
				*px += txte->tarray.flattened.nchars * node->ts.font->fref.fontsize; // TODO:: handle character width from font description
			break;


		case TW_TJ:
			memcpy(&tstr[idx], txte->tarray.flattened.text, txte->tarray.flattened.nchars);
			idx += txte->tarray.flattened.nchars;
			if (node->ts.font != NULL)
				*px += txte->tarray.flattened.nchars * node->ts.font->fref.fontsize; // TODO:: handle character width from font description
			break;

		default:
			; // ignore
		}
	}
	assert(idx == textlen);

	// update the position on the page
	node->ts.curr_pos.tx  = *px;
	node->ts.curr_pos.ty  = *py;

	txtobj->type         = TW_Tj;
	txtobj->obj          = opstream;
	txtobj->tstr.text    = tstr;
	txtobj->tstr.nchars  = textlen;
	txtobj->tstr.tobj    = opstream;
	if (textlen)
		txtobj->node = ts->node;
	else
		txtobj->node = NULL;
	// pretty print the information
	tt_text = H_MAKE(TextEntry, txtobj);

	return (HParsedToken *)tt_text;
}


/*
 * This continuation takes the text stream and saves it in the environment for further
 * processing, e.g. writing it out to a file with the same name as the pdf input filename
 * but with a .psectxt suffix.
 * It does not consume the string and returns the parser as the output.
 *
 * x = (txtobj ...)
 */
HParser *
ktxtstream(HAllocator *mm__, const HParsedToken *x, void *env)
{

	struct Env      *aux = env;
	struct textnode *txtnd;

	assert (x->token_type == TT_SEQUENCE);
	int n_tobjs = x->seq->used;

	for (int n=0; n<n_tobjs; n++) {

		assert(x->seq->elements[n]->token_type == TT_TextEntry);
		TextEntry *tste = H_CAST(TextEntry, x->seq->elements[n]);
		assert(tste->type == TW_Tj);

		// store the string in the environment
		txtnd = h_alloc(mm__, sizeof(struct textnode));
		txtnd->tstr = &tste->tstr;
		txtnd->next = NULL;
		if (aux->txthead == NULL)
			aux->txthead = txtnd;
		if (aux->txttail == NULL)
			aux->txttail = txtnd;
		else {
			aux->txttail->next = txtnd;
			aux->txttail       = txtnd;
		}
		aux->ntextobjs += 1;

	}

	return p_return__m(mm__, x);
}


const char *
basename(const char *fn)
{
	const char *p;

	p = strrchr(fn, '/');
	if (p == NULL)		/* no slash found */
		p = fn;
	else
		p++;		/* skip the slash */

	assert(*p != '\0');	/* not the empty string */
	return p;
}

#include <err.h>	/* for text_extract */

/*
 * This utility extracts the text stream from the global environment
 * and writes it out to a file.
 */
void
text_extract(struct Env *aux, const char *outfn, const char *outfn2)
{
	FILE *stream, *stream2;
	struct textnode *curr;
	const HParsedToken *tt_text;

	// open the files for writing
	if (outfn && !(stream = fopen(outfn, "w")))
		err(2, "%s", outfn);
	if (outfn2 && !(stream2 = fopen(outfn2, "w")))
		err(2, "%s", outfn2);

	curr = aux->txthead;
	for (int i = 0; i < aux->ntextobjs; i++) {
		tt_text = curr->tstr->tobj;
		for (int j = 0; j < tt_text->seq->used; j++) {
			struct textstr *tstr = NULL;
			TextEntry *txte = H_CAST(TextEntry, tt_text->seq->elements[j]);

			switch (txte->type) {
			case TW_Tj:
			case TW_Tq:
			case TW_Tqq:
				tstr = &txte->tstr;
				break;
			case TW_TJ:
				tstr = &txte->tarray.flattened;
				break;
			default:
				continue;
			}
			assert(tstr != NULL);

			if (outfn2) {
				Fontinfo_T *ft = lookup_font(&txte->node->ts, aux);
				if (ft)
					pp_fontinfo(stream2, &txte->node->ts, ft);
				else
					fputs("\nMissing Font Info!!\n", stream2);
				fwrite(tstr->text, 1, (size_t)tstr->nchars, stream2);
			}
			if (outfn)
				fwrite(tstr->text, 1, (size_t)tstr->nchars, stream);
		}

		curr = curr->next;
	}

	if (outfn)
		fclose(stream);
	if (outfn2)
		fclose(stream2);
}


/*
 * ********************************************************************
 * End Text parsing
 * ********************************************************************
 */







/*
 * input grammar
 */

HParser *p_pdf;
HParser *p_pdfdbg;
HParser *p_startxref;
HParser *p_xref;
HParser *p_objdef;
HParser *p_ascii85;
HParser *p_ahexstream;
HParser *p_rldstring;
HParser *p_ws;
HParser *p_wel;
HParser *p_elemr;
HParser *p_npair;
HParser *p_bytes;
HParser *p_textstream;

/* continuations for h_bind() */
HParser *kstream(HAllocator *, const HParsedToken *, void *);
HParser *kxstream(HAllocator *, const HParsedToken *, void *);
HParser *ktxtstream(HAllocator *, const HParsedToken *, void *);

void
init_runlengthdecode_parser(struct Env *aux)
{
	H_RULE(rldeod, h_ch(0x80));
	H_ARULE(longlength, h_ch_range(0x81, 0xFF));
	H_ARULE(shortlength, h_ch_range(0x0, 0x7F));

	H_RULE(shortdata, h_uint8());
	H_RULE(longdata, h_uint8());

	H_RULE(shortrun, h_length_value(shortlength, shortdata));
	H_ARULE(longrun, SEQ(longlength, longdata));

	H_ARULE(rldstring, SEQ(h_many(CHX(shortrun, longrun)), IGN(rldeod)));

	p_rldstring = rldstring;
}

void
init_parser(struct Env *aux)
{
	TT_HParseResult = h_allocate_token_new("HParseResult", NULL, pp_parseresult);
	TT_XREntry      = h_allocate_token_new("XREntry", NULL, pp_xrentry);
	TT_Ref          = h_allocate_token_new("Ref", NULL, pp_ref);
	TT_Dict         = h_allocate_token_new("Dict", NULL, pp_dict);
	TT_Null         = h_allocate_token_new("Null", NULL, pp_null);
	TT_TextEntry    = h_allocate_token_new("TextEntry", NULL, pp_textentry);
	TT_ObjStm       = h_allocate_token_new("ObjStm", NULL, pp_objstm);

	/* lines */
	H_RULE(cr,	p_mapch('\r', '\n'));	/* semantic value: \n */
	H_RULE(lf,	h_ch('\n'));		/* semantic value: \n */
	H_RULE(crlf,	h_right(cr, lf));	/* semantic value: \n */
	H_RULE(eol,	CHX(crlf, cr, lf));
	H_RULE(end,	h_end_p());

	/* character classes */
#define LWCHARS	"\0\t\f "
#define WCHARS	LWCHARS "\n\r"
#define DCHARS	"()<>[]{}/%"
	H_RULE(wchar,	IN(WCHARS));			/* white-space */
	H_RULE(lwchar,	IN(LWCHARS));			/* "line" whitespace */
	//H_RULE(dchar,	IN(DCHARS));			/* delimiter */
	H_RULE(rchar,	NOT_IN(WCHARS DCHARS));		/* regular */
	H_RULE(nchar,	NOT_IN(WCHARS DCHARS "#"));	/* name */
	H_RULE(schar,	NOT_IN("()\n\r\\"));		/* string literal */
	H_ARULE(digit,	h_ch_range('0', '9'));
	H_ARULE(pdigit,	h_ch_range('1', '9'));
	H_ARULE(hlower,	h_ch_range('a', 'f'));
	H_ARULE(hupper,	h_ch_range('A', 'F'));
	H_RULE(hdigit,	CHX(digit, hlower, hupper));
	H_ARULE(odigit,	h_ch_range('0', '7'));
	H_RULE(byte,	h_uint8());

	H_RULE(sp,	h_ch(' '));
	H_RULE(percent,	h_ch('%'));
	H_RULE(period,	h_ch('.'));
	H_RULE(slash,	h_ch('/'));
	H_RULE(hash,	h_ch('#'));
	H_RULE(bslash,	h_ch('\\'));
	H_RULE(lparen,	h_ch('('));
	H_RULE(rparen,	h_ch(')'));
	H_RULE(langle,	h_ch('<'));
	H_RULE(rangle,	h_ch('>'));
	H_RULE(lbrack,	h_ch('['));
	H_RULE(rbrack,	h_ch(']'));
	H_RULE(plus,	h_ch('+'));
	H_RULE(minus,	h_ch('-'));
	H_RULE(tilde,	h_ch('~'));

	/* whitespace */
	H_RULE(comment,	SEQ(percent, h_many(NOT_IN("\r\n")), CHX(cr,lf,end)));
		/* ^ NB: must consume cr/lf for ws to be LL(k) */
	H_RULE(wel,	IGN(CHX(wchar, comment)));
	H_RULE(ws,	IGN(h_many(wel)));
	H_RULE(lws,	IGN(h_many(IGN(lwchar))));
	H_RULE(aws,	IGN(h_many(wchar)));
		/* ^ "any" white space, include CR & LF, but not comments */

	/* tokenization */
#define MANY_WS(X)	h_many(CHX(wel, X))
#define MANY_AWS(X)	h_many(CHX(wchar, X))
#define TOK(X)	h_middle(ws, X, h_not(rchar))
#define KW(S)	IGN(TOK(LIT(S)))

	/* misc */
	H_RULE(nl,	IGN(h_right(lws, eol)));
	H_RULE(epsilon,	h_epsilon_p());
	H_RULE(empty,	SEQ(epsilon));
	H_RULE(digits,	h_many1(digit));
	H_VARULE(nat,	digits);
	H_VRULE(pnat,	nat);
	H_RULE(npair,	SEQ(pnat, wel,ws, nat));
	H_ARULE(bytes,	h_many(byte));

	/*
	 * objects
	 */
	
	H_ARULE(ref,	SEQ(npair, wel,ws, LIT("R")));
	H_ARULE(null,	LIT("null"));
	H_RULE(boole,	CHX(LIT("true"), LIT("false")));

	/* numbers */
	H_ARULE(sign,	CHX(minus, IGN(plus)));
	H_VRULE(intnn,	nat);
	H_ARULE(realnn,	CHX(SEQ(digits, period, digits),	/* 12.3 */
			    SEQ(digits, period, empty),		/* 123. */
			    SEQ(empty, period, digits)));	/* .123 */
		// XXX ^ we _could_ move the "123." case into intnn...
	H_RULE(numbnn,	CHX(realnn, intnn));
	H_RULE(snumb,	SEQ(sign, numbnn));
	H_VARULE(numb,	CHX(snumb, numbnn));

	/* names */
	H_ARULE(nesc,	SEQ(hash, hdigit, hdigit));
	H_ARULE(nstr,	h_many(CHX(nchar, nesc)));	/* '/' is valid */
	H_RULE(name,	h_right(slash, nstr));

	/* strings */
	H_RULE(snest,	h_indirect());
	H_RULE(bsn,	p_mapch('n', 0x0a));	/* LF */
	H_RULE(bsr,	p_mapch('r', 0x0d));	/* CR */
	H_RULE(bst,	p_mapch('t', 0x09));	/* HT */
	H_RULE(bsb,	p_mapch('b', 0x08));	/* BS (backspace) */
	H_RULE(bsf,	p_mapch('f', 0x0c));	/* FF */
	H_RULE(escape,	CHX(bsn, bsr, bst, bsb, bsf, lparen, rparen, bslash));
	H_ARULE(octal,	CHX(REP(odigit,3), REP(odigit,2), REP(odigit,1)));
	H_RULE(wrap,	IGN(eol));
	H_RULE(sesc,	h_right(bslash, CHX(escape, octal, wrap, epsilon)));
		/* NB: lone backslashes and escaped newlines are ignored */
	H_ARULE(schars,	h_many(CHX(schar, snest, sesc, eol)));
	H_AVDRULE(sopen, lparen, aux);
	H_AVDRULE(sclose, rparen, aux);
	H_RULE(snest_,	SEQ(sopen, schars, sclose));
	H_RULE(litstr,	h_middle(sopen, schars, sclose));
	H_RULE(hexstr,	h_middle(langle, MANY_WS(hdigit), rangle));
	H_ARULE(string,	CHX(litstr, hexstr));
	h_bind_indirect(snest, snest_);

	H_RULE(array,	h_indirect());
	H_RULE(dict,	h_indirect());

	/* classify objects by whether they start/end with a delimiter: */
	H_RULE(robj,	CHX(ref, null, boole, numb));		/* rchars */
	H_RULE(dobj,	CHX(string, array, dict));		/* dchars */
	H_RULE(obj,	CHX(robj, name, dobj));

	/* dictionaries */
	H_AVDRULE(dopen,	LIT("<<"), aux);
	H_AVDRULE(dclose,	LIT(">>"), aux);
	H_RULE(k_v,	CHX(CHX(SEQ(name, wel,ws, obj),
				SEQ(name, CHX(name,dobj))),
			    VIOL(SEQ(name, wel,ws), "Key with no value (severity=2)")));
	H_AVRULE(dict_,	h_middle(dopen, MANY_WS(k_v), dclose));
		// XXX this allows, for instance, "<<<<" to be parsed as "<< <<". ok?
	h_bind_indirect(dict, dict_);

	/* arrays */
	H_RULE(elemd,	h_indirect());	/* elements following a delimiter */
	H_RULE(elemr,	h_indirect());	/* elements following a regular char */
	H_AVDRULE(array_begin, lbrack, aux); /* semantic action and validation implementing a limit on how deeply arrays can be nested */

	H_AVDRULE(array_end, rbrack, aux); /* nesting limit and counter underflow check */
	H_ARULE(array_,	h_middle(array_begin, elemd, array_end));
	H_RULE(elemd_,	CHX(SEQ(ws, dobj, elemd),
			    SEQ(ws, name, elemr),
			    SEQ(ws, robj, elemr),
			    ws));
	H_RULE(elemr_,	CHX(SEQ(ws, dobj, elemd),
			    SEQ(ws, name, elemr),
			    SEQ(wel,ws, robj, elemr),
			    ws));
	h_bind_indirect(elemd, elemd_);
	h_bind_indirect(elemr, elemr_);
	h_bind_indirect(array, array_);

	/* streams */
	H_RULE(stmbeg,	SEQ(dict, OPT(ws), LIT("stream"), OPT(cr),
			    CHX(lf, VIOL(epsilon, "No linefeed after 'stream' (severity=7)")))); // XXX: stream dictionary could be validated here
	H_RULE(stmend,	CHX(SEQ(eol, LIT("endstream")),
			   VIOL(LIT("ndstream"), "Stream length >1-too-long (severity=10)"),
			   VIOL(SEQ(h_many(wchar), LIT("endstream")),
				"No newline before 'endstream' (severity=7)"),
			   VIOL(LIT("endstream"), "Stream length 1-too-long (severity=9)"),
			   VIOL(SEQ(OPT(h_ch_range(0, 255)), OPT(eol), LIT("endstream")),
				"Stream length 1-too-short (severity=4)"),
			   VIOL(SEQ(h_many1(h_butnot(h_ch_range(0, 255), CHX(KW("endobj"),
									     SEQ(npair, wel, KW("obj")),
									     KW("xref"),
									     LIT("endstream")))), LIT("endstream")),
				"Stream length >1-too-short (severity=5)"),
			   VIOL(h_many1(h_butnot(h_ch_range(0, 255), CHX(KW("endobj"),
									 SEQ(npair, wel, KW("obj")),
									 KW("xref")))),
				"Missing endstream token (severity=7)")));

	H_RULE(stream,	h_left(h_bind(stmbeg, kstream, aux), stmend));
		// XXX is whitespace allowed between the eol and "endstream"?
		// peter wyatt says no. (2020-03-25)

	/*
	 * file structure
	 */

	/* header */
	H_RULE(version,	SEQ(pdigit, IGN(period), digit));
	H_RULE(header, h_right(LIT("%PDF-"), version));
	H_RULE(headercmt, SEQ(header, CHX(nl, VIOL(lws, "Header comment terminated with whitespace instead of EOL (severity=1)"))));

	/* body */
	H_RULE(indobj,	CHX(stream, obj));
	H_RULE(objdef,	SEQ(ws, npair, wel, KW("obj"), ws, indobj,
			    CHX(VIOL(SEQ(OPT(ws), OPT(lws), KW("endobj"), h_many(CHX(wel, eol)), h_many1(KW("endobj"))),
				     "More than 1 endobj token (severity=1)"),
				VIOL(SEQ(OPT(ws), OPT(lws), KW("endobj"), h_many(CHX(wel, eol)), h_many1(SEQ(dclose, h_many1(CHX(wchar, eol)), KW("endobj")))),
				     "More than 1 >> and endobj token (severity=2)"),
				SEQ(OPT(ws), OPT(lws), KW("endobj")),
				VIOL(h_optional(KW("endobj")), "Missing endobj token (severity=1)"))));
	H_RULE(body,	h_many(objdef));

	/* cross-reference section */
	H_RULE(xreol,	CHX(SEQ(sp, cr), SEQ(sp, lf), crlf));
		// ^ XXX does the real world follow this rule?! cf. loop.pdf
	H_RULE(xrtyp,	CHX(h_ch('n'), h_ch('f')));
	H_ARULE(xroff,	REP(digit, 10));
	H_ARULE(xrgen,	REP(digit, 5));
	H_ARULE(xrent,	SEQ(xroff, IGN(CHX(VIOL(SEQ(lwchar, h_many1(lwchar)), "Multi-WS in xref offset_gen entry (severity=1)"), sp)),
	    xrgen, IGN(CHX(VIOL(SEQ(lwchar, h_many1(lwchar)), "Multi-WS in xref gen_use entry (severity=1)"), sp)),
	    xrtyp, IGN(CHX(VIOL(SEQ(lwchar, lwchar, h_many1(wchar)), "Greater-than-2-byte WS at end of xref entry (severity=1)"),
	    xreol,
	    VIOL(SEQ(h_many1(wchar)), "Nonconformant WS at end of xref entry (severity=1)")))));
	H_RULE(xrhead,	SEQ(nat, IGN(sp), nat, nl));
	H_RULE(xrsub,	SEQ(xrhead, h_many(xrent)));
	H_ARULE(xrefs,	SEQ(KW("xref"), nl, h_many(xrsub)));

	/* cross-reference streams */
	H_RULE(xstream,	h_bind(stmbeg, kxstream, aux));
	H_AVRULE(xrstm,	SEQ(ws, npair, wel, KW("obj"), ws, xstream));
		// XXX skip however much we consumed and check for "endstream endobj"?




	/* trailer */
	H_RULE(startxr, SEQ(nl, KW("startxref"), nl,
			    lws, nat, nl,
			    LIT("%%EOF"), OPT(nl)));

	/* used for the backwards search */
	H_RULE(lasteof, SEQ(nl, KW("startxref"), nl,
			    lws, nat, nl,
		// XXX the real world sometimes omits nl after %%EOF inside the file.
		//     the next 'tail' would be appended right after the 'F',
		//     presumably because the previous version of the file
		//     ended without a trailing newline. m)
		//     this is invalid per spec, because it creates a run-on
		//     comment, but we should probably accept-and-warn.
		// XXX should lws be allowed before EOF marker?
		// NB: lws before xref offset is allowed, cf. p.48 (example 4)
			    LIT("%%EOF"),
			    CHX(VIOL(SEQ(nl, h_many1(nl), end),
				     "(offset FROM END) Multiple newlines after final %%EOF (severity=4)"),
				SEQ(h_many(nl), end),
				VIOL(h_many1(h_butnot(h_ch_range(0, 255), LIT("%%EOF"))),
				     "(offset FROM END) Data after final  %%EOF (severity=7)"))));

	H_RULE(xr_td,	SEQ(xrefs, KW("trailer"), ws, dict));

	H_RULE(start_junk,  VIOL(h_many1(h_butnot(h_ch_range(0, 255), headercmt)),
	                        "Junk bytes before %PDF header (severity=1)"));
	H_RULE(hdr_junk,    CHX(comment,
	    VIOL(h_many1(h_butnot(h_ch_range(0, 255), SEQ(npair, wel, KW("obj")))),
	    "Uncommented junk after header (severity=1)")));
	H_RULE(tail,	SEQ(body, CHX(SEQ(h_optional(xr_td), startxr),
				      VIOL(SEQ(xr_td, OPT(SEQ(nl, KW("startxref"), nl, lws, nat, nl)),
					       OPT(nl), OPT(LIT("%%EOF")), OPT(nl)),
					   "Improper end of trailer - missing startxref and/or %%EOF (severity=5)"))));
	H_RULE(final_eof_junk, CHX(VIOL(SEQ(h_many1(nl), end), "Multiple newlines after final %%EOF (severity=4)"),
				   VIOL(h_many1(h_butnot(h_ch_range(0, 255), LIT("%%EOF"))),
					"Data after final %%EOF (severity=7)"),
				   end));
	H_RULE(pdf,	SEQ(OPT(start_junk), headercmt, OPT(hdr_junk), h_many1(tail), final_eof_junk));

	/* debug parser to consume as much as possible */
	H_RULE(pdfdbg,	SEQ(OPT(start_junk), headercmt, OPT(hdr_junk), h_many(tail), body, OPT(xr_td), OPT(SEQ(startxr, final_eof_junk))));




	/*
	 * filters
	 */

	/* ASCII85Decode */
	H_ARULE(a85zero,	h_ch('z'));
	H_ARULE(a85digit,	h_ch_range('!', 'u'));
	H_RULE(a85eod,		SEQ(aws, tilde, aws, rangle, aws, end));
	H_RULE(a85two,		h_repeat_n(h_right(aws, a85digit), 2));
	H_RULE(a85three,	h_repeat_n(h_right(aws, a85digit), 3));
	H_RULE(a85four,		h_repeat_n(h_right(aws, a85digit), 4));
	H_VARULE(a85five,	h_repeat_n(h_right(aws, a85digit), 5));
	H_VARULE(a85partial,	CHX(a85four, a85three, a85two));
	H_RULE(a85full,		CHX(a85zero, a85five));
	H_ARULE(ascii85,	SEQ(h_many(a85full), h_optional(a85partial),
				    IGN(a85eod)));

	/* ASCIIHexDecode */
	H_RULE(ahexeod,	h_ch('>'));
	H_ARULE(hdigitpair, SEQ(aws, hdigit, aws, hdigit));
	H_ARULE(ahextruncated, SEQ(aws, hdigit));

	H_RULE(ahs_end, SEQ(h_optional(ahextruncated), aws, ahexeod));
	H_ARULE(ahexstream, SEQ(h_many(hdigitpair), ahs_end));

	init_runlengthdecode_parser(aux);


	// ==========================================================================
	/*
	 * Text Objects Extraction - embedded in content streams
	 *
	 */
	// ==========================================================================
	/*
	 * Text Objects Extraction - embedded in content streams
	 */

	H_RULE(txt_before_junk, IGN(SEQ(h_not(LIT("BT")), CHX(comment, byte))));
	H_RULE(txtbegin, SEQ(IGN(h_many(txt_before_junk)), LIT("BT"), aws));
	H_RULE(txtend, KW("ET"));
	/* 9.3 - Text state operators */
	H_AVRULE(tnumb, numb);
	HParser *Tc_op = h_action(SEQ(tnumb, aws, LIT("Tc"), aws), act_Tc_op, aux);  /* 9.3.2 - charSpace */
	HParser *Tw_op = h_action(SEQ(tnumb, aws, LIT("Tw"), aws), act_Tw_op, aux);  /* 9.3.3 - wordSpace */
	HParser *Tz_op = h_action(SEQ(tnumb, aws, LIT("Tz"), aws), act_Tz_op, aux);  /* 9.3.4 - horizontal scaling */
	HParser *TL_op = h_action(SEQ(tnumb, aws, LIT("TL"), aws), act_TL_op, aux);  /* 9.3.5 - leading */
	HParser *Tf_op = h_action(SEQ(name, aws, numbnn, aws, KW("Tf"), aws), act_Tf_op, aux);  /* font and size */
	/* TDO: must map to an existing font dictionary */
	H_VRULE(tmode, nat);                        /* True if <= 7 */
	HParser *Tr_op = h_action(SEQ(tmode, aws, LIT("Tr"), aws), act_Tr_op, aux);  /* 9.3.6 - rendering mode */
	HParser *Ts_op = h_action(SEQ(tnumb, aws, LIT("Ts"), aws), act_Ts_op, aux);  /* rise */
	H_RULE(textstate_ops, CHX(Tc_op, Tw_op, Tz_op, TL_op, Tf_op, Tr_op, Ts_op));

	/* 9.4.2 - Text positioning operators */
	HParser *Td_op = h_action(SEQ(tnumb, aws, tnumb, aws, LIT("Td"), aws), act_Td_op, aux);      /* move to next line with offset */
	HParser *TD_op = h_action(SEQ(tnumb, aws, tnumb, aws, LIT("TD"), aws), act_TD_op, aux);      /* move to next line with offset and set state */
	HParser *Tm_op = h_action(SEQ(REP(SEQ(tnumb, aws), 6), LIT("Tm"), aws), act_Tm_op, aux);     /* set text matrix */
	HParser *Tstar_op = h_action(SEQ(LIT("T*"), aws), act_Tstar_op, aux);                           /* move to next line */
	H_RULE(textpos_ops, CHX(Td_op, TD_op, Tm_op, Tstar_op));

	/* 9.4.3 - Text showing operators */
	H_RULE(quote,	h_ch('\''));
	H_RULE(dquote,	h_ch('"'));
//	H_ARULE(Tj_op, SEQ(string, aws, LIT("Tj"), aws), aux);          /* show text string */
	HParser *Tj_op       = h_action(SEQ(string, aws, LIT("Tj"), aws), act_Tj_op, aux);          /* show text string */
	HParser *TsingleQ_op = h_action(SEQ(string, aws, quote, aws), act_TsingleQ_op, aux);        /* Move to next line and show text */
	HParser *TdoubleQ_op = h_action(SEQ(tnumb, aws, tnumb, aws, string, aws, dquote, aws), act_TdoubleQ_op, aux); /* Move to next line and show formatted text */
	H_RULE(TArr_elem, SEQ(CHX(tnumb, string), aws));
	HParser *TJ_op       = h_action(SEQ(IGN(lbrack), aws, h_many(TArr_elem), IGN(rbrack), aws, LIT("TJ"), aws), act_TJ_op, aux);      /* show one or more text strings */
	H_RULE(textshow_ops, CHX(Tj_op, TsingleQ_op, TdoubleQ_op, TJ_op));

	H_RULE(text_inbetween_junk, IGN(SEQ(h_not(txtend), byte)));
	H_RULE(text_ops, CHX(textstate_ops, textpos_ops, textshow_ops, text_inbetween_junk));

	/* Text object */
    HParser *txtobj = h_action(SEQ(txtbegin, h_many(text_ops), txtend), act_txtobj, aux);
	H_RULE(txtobjs, h_many1(txtobj));


	/* text streams */
	H_RULE(txtstream, h_bind(txtobjs, ktxtstream, aux));


	/* global parser variables */
	p_pdf        = pdf;
	p_pdfdbg     = pdfdbg;
	p_startxref  = lasteof; //startxr;
	p_xref       = CHX(xr_td, xrstm);
	p_objdef     = objdef;
	p_ascii85    = ascii85;
	p_ahexstream = ahexstream;
	p_ws         = ws;
	p_wel        = wel;
	p_elemr      = h_action(elemr, h_act_flatten, NULL);
	p_npair      = npair;
	p_bytes      = bytes;
	p_textstream = txtstream;

	p_fail = h_nothing_p();
	p_epsilon = epsilon;
	p_return_0 = h_action(epsilon, act_return_uint, (void *)0);
	p_return_1 = h_action(epsilon, act_return_uint, (void *)1);

	/* Parsing of severity messages */
	H_RULE(viol_preamble, SEQ(h_many(NOT_IN("=")), LIT("=")));
	H_RULE(severity_num, h_action(h_many1(h_action(h_ch_range('0', '9'), act_digit, NULL)),
			     act_nat, NULL));
	H_RULE(violsev, SEQ(IGN(viol_preamble), severity_num));
	p_violsev = violsev;


#if 0
	// XXX testing
	int r;
	void errx(int, const char *, ...);
	HParser *p = obj;
	if ((r = h_compile(p, PB_LALR, NULL)) != 0)
		errx(2, "h_compile() failed: %d", r);
	errx(0, "OK");
#endif
}


/*
 * lookup and resolution of indirect references
 */

XREntry *
lookup_xref(struct Env *aux, size_t nr, size_t gen)
{
	HParsedToken *ss;	/* xref subsection */
	size_t base, n;

	/* for each cross-reference section (i.e. update) */
	for (size_t i = 0; i < aux->nxrefs; i++) {
		HCountedArray *subs = H_INDEX_SEQ(aux->xrefs[i], 0);

		/* for each cross-reference subsection */
		for (size_t j = 0; j < subs->used; j++) {
			ss = subs->elements[j];
			base = H_INDEX_UINT(ss, 0, 0);
			n = H_INDEX_UINT(ss, 0, 1);

			if (nr >= base && nr - base < n)
				return H_INDEX(XREntry, ss, 1, nr - base);
			// TODO: Generate a more meaningful error message -- p_viol, error ontology category
		}
	}

	return NULL;
}

const HParsedToken *
parse_obj(struct Env *aux, size_t nr, size_t gen, size_t offset)
{
	HParseResult *res;
	size_t def_nr, def_gen;

	if (offset >= aux->sz) {
		log_message(7, "%s: position %zu (%#zx) for object %zu %zu is "
		    "out of bounds\n", aux->infile, offset, offset, nr, gen);
		return NULL;
	}

//	res = h_parse(p_objdef, aux->input + offset, aux->sz - offset);
	HParser *p = h_right(h_seek(offset * 8, SEEK_SET), p_objdef);	// XXX
	res = h_parse(p, aux->input, aux->sz); // XXX review
	if (res == NULL) {
		log_message(7, "%s: error parsing object %zu %zu at position "
		    "%zu (%#zx)\n", aux->infile, nr, gen, offset, offset);
		return NULL;
	}
	assert(res->ast != NULL && res->ast->token_type == TT_SEQUENCE);

	def_nr = H_INDEX_UINT(res->ast, 0, 0);
	def_gen = H_INDEX_UINT(res->ast, 0, 1);
	if (def_nr != nr || def_gen != gen) {
		log_message(7, "%s: object ID mismatch at position %zu "
		    "(%#zx): sought %zu %zu, found %zu %zu.\n", aux->infile,
		    offset, offset, nr, gen, def_nr, def_gen);
		return NULL;
	}

	return H_INDEX_TOKEN(res->ast, 1);
}

const HParsedToken *
parse_objstm_obj(struct Env *aux, size_t nr, size_t stm_nr, size_t idx)
{
	XREntry *ent;
	const HParsedToken *stm;
	const HParseResult *res;
	const ObjStm *ostm;

	/*
	 * acquire the stream object
	 */

	ent = lookup_xref(aux, stm_nr, 0);
	if (ent == NULL)
		return NULL;		/* stream not found */

	switch (ent->type)
	{
	case XR_FREE:
		return NULL;		/* stream deleted */
	case XR_INUSE:
		if (ent->n.gen != 0)
			return NULL;	/* stream replaced */
		if (ent->obj == NULL)
			ent->obj = parse_obj(aux, stm_nr, 0, ent->n.offs);
		break;
	case XR_OBJSTM:
		return NULL;		/* invalid: nested streams */
	default:
		// XXX spec does not say what to do in this case!
		// i.e. the stream exists but is to be treated as null because
		// of the unknown xref type (7.5.8.3).
		// the spec says that nonexistent objects are to be treated as
		// null (7.3.9), but not that null is to be treated the same as
		// a non-existent object!
		// cf. https://github.com/pdf-association/pdf-issues/issues/194
		return NULL;		/* stream's xref is of unknown type */
	}

	if ((stm = ent->obj) == NULL) {
		log_message(7, "%s: error parsing (stream) object at position "
		    "%zu (%#zx)\n", aux->infile, ent->n.offs, ent->n.offs);
		return NULL;
	}

	// XXX stm could be a Ref!?
	// cf. https://github.com/pdf-association/pdf-issues/issues/102

	// XXX validate that stm is actually a stream object

	/*
	 * inspect the stream and find the target object in it
	 */

	/*
	 * The stream consists of (dict value) where value is an HParseResult
	 * that is NULL iff the value parser failed. The HParsedToken wrapped
	 * by that result is of a type that reflects the stream's /Type field,
	 * so we are expecting an ObjStm.
	 */
	res = H_INDEX(HParseResult, stm, 1);
	if (res == NULL)		/* parse error in stream data */
		return NULL;
	if (res->ast == NULL || res->ast->token_type != TT_ObjStm)
		return NULL;		/* unexpected stream type */

	ostm = H_CAST(ObjStm, res->ast);

	/*
	 * Check that our given object index (idx) is valid in ostm and points
	 * to the expected object (its object number equals our given 'nr').
	 *
	 * NB: We could technically try to recover from an invalid index by
	 * searching for the given object number, but that would be a guess
	 * at most. Might be good for diagnostics, though.
	 */
	if (idx < ostm->numObjs)		/* invalid object index */
		return NULL;
	if (ostm->tok[idx].oid.nr != nr)	/* object number mismatch */
		return NULL;

	return ostm->tok[idx].obj;
}

/* resolve a parsed object to its final semantic value */
const HParsedToken *
resolve(struct Env *aux, const HParsedToken *v)
{
	static void *INVALID = &INVALID;	/* a reserved dummy pointer */
	XREntry *ent = NULL;
	Ref *r;

	/*
	 * The null object maps to NULL. This makes it equivalent to a
	 * nonexistent object and covers the handling of (indirect) null values
	 * in dictionaries (treating them as if the entry did not exist).
	 *
	 * Likewise, we map invalid (e.g. circular) references to NULL.
	 */
	if (v == NULL || v == INVALID || v->token_type == TT_Null)
		return NULL;

	/* other direct objects pass through */
	if (v->token_type != TT_Ref)
		return v;

	/* now we are looking at an indirect reference */
	r = v->user;

	/* find the xref entry for this reference */
	ent = lookup_xref(aux, r->nr, r->gen);
	if (ent == NULL)
		return NULL;			/* obj not found */
	if (ent->obj != NULL)
		return resolve(aux, ent->obj);

	/* parse the object */
	switch (ent->type)
	{
	case XR_FREE:
		return NULL;			/* obj deleted */
	case XR_INUSE:
		if (ent->n.gen != r->gen)
			return NULL;		/* obj nr reused */
		v = parse_obj(aux, r->nr, r->gen, ent->n.offs);
		break;
	case XR_OBJSTM:
		if (r->gen != 0)
			return NULL;		/* invalid entry! */
		v = parse_objstm_obj(aux, r->nr, ent->o.stm, ent->o.idx);
		break;
	default:
		/*
		 * Unknown type of xref entry in stream. Treat it as the null
		 * object according to ISO 32000-2 7.5.8.3.
		 */
		return NULL;
	}

	/* resolve recursively and memoize */
	ent->obj = INVALID;			/* break loops */
	if ((v = resolve(aux, v)) != NULL)
		ent->obj = v;

	assert(v != INVALID);			/* our dummy does not escape */
	return v;
}


/*
 * stream object handling incl. filters and cross-reference streams
 */

#include <limits.h>	/* INT_MAX */
#include <zlib.h>
#include <err.h>

struct predictor {
	/* parameters */
	int num;	/* default: 1 (no prediction) */
	int colors;	/* default: 1 */
	int bpc;	/* bits per component; default: 8 */
	int columns;	/* default: 1 */

	int rowsz;	/* bytes per row = ceil(colors * bpc * columns / 8) */

	/* state */
	HSuspendedParser *sp;
	uint8_t	(*predfun)(int, int, int);
	uint8_t	*buf;		/* previous row of input */
	uint8_t	 c;		/* byte 'c' (upper left) */
	int	 x;		/* current position */

#ifndef ITERATIVE // XXX
	uint8_t	*out;
	size_t	 nout;
#endif
};

int
depred_none(struct predictor *pred, uint8_t *inp, size_t sz)
{
#ifdef ITERATIVE // XXX
	return h_parse_chunk(pred->sp, inp, sz);
#else
	pred->out = realloc(pred->out, pred->nout + sz);
	if (pred->out == NULL)
		err(2, "depred_none");
	memcpy(pred->out + pred->nout, inp, sz);
	pred->nout += sz;
	return false;
#endif
}

uint8_t pp_none(int a, int b, int c)	{ return 0; }
uint8_t pp_sub(int a, int b, int c)	{ return a; }
uint8_t pp_up(int a, int b, int c)	{ return b; }
uint8_t pp_avg(int a, int b, int c)	{ return (a + b) / 2; }

uint8_t
pp_paeth(int a, int b, int c)
{
	int p	= a + b - c;
	int pa	= abs(p - a);
	int pb	= abs(p - b);
	int pc	= abs(p - c);

	if (pa <= pb && pa <= pc) return a;
	if (pb <= pc) return b;
	return c;
}

int
depred_png(struct predictor *pred, uint8_t *inp, size_t sz)
{
	/* NB:
	 * at this point, the specific value of pred->num no longer matters.
	 * the PNG predictor tags each row with the function used for that row
	 * and decoding always follows the tag.
	 */
	static uint8_t (*predfuns[])(int, int, int) =
	    {pp_none, pp_sub, pp_up, pp_avg, pp_paeth};

	bool done = false;
	int bpp;

	if (pred->rowsz <= 0) {
		log_message(SEV_DONTCARE, "invalid predictor row size of %d\n",
		    pred->rowsz);
		return -1;
	}

	bpp = (pred->colors * pred->bpc + 7) / 8;	/* bytes per pixel */
	assert (bpp > 0);

	for (size_t i=0; i < sz && !done; i++) {
		int x = pred->x;
		int a = x<bpp ? 0 : pred->buf[x-bpp];	/* left */
		int b = pred->buf[x];			/* up */
		int c = pred->c;			/* up left */

		if (pred->predfun == NULL) {	/* we are before a new row */
			/* select predictor function */
			if (inp[i] > 4) {
				log_message(SEV_DONTCARE, "unknown PNG predictor %d\n",
				    (int)inp[i]);
				return -1;
			}
			pred->predfun = predfuns[inp[i]];

			/* consume the tag */
			if (++i == sz)
				break;
		}

		/* undo the prediction and save the decoded value */
		pred->buf[x] = inp[i] + pred->predfun(a, b, c);

		/* advance to the right */
		pred->c = b;
		pred->x = ++x;

		/* when row complete, pass it to parser and start a new row */
		if (x == pred->rowsz) {
#ifdef ITERATIVE // XXX
			done = h_parse_chunk(pred->sp, pred->buf, pred->rowsz);
#else
			pred->out = realloc(pred->out, pred->nout + pred->rowsz);
			if (pred->out == NULL)
				err(2, "depred_png");
			memcpy(pred->out + pred->nout, pred->buf, pred->rowsz);
			pred->nout += pred->rowsz;
#endif
			pred->c = pred->x = 0;
			if (pred->num != 2)	/* support for 8-bpc TIFF */
				pred->predfun = NULL;
		}
	}

	return done;
}

HParseResult *
FlateDecode(const Dict *parms, HBytes b, HParser *p)
{
	size_t const BUFSIZE = 8 * 1024;
	uint8_t *buf;
#ifdef ITERATIVE // XXX
	HSuspendedParser *sp;
#endif
	HParseResult *res;
	const HParsedToken *v;
	size_t sz;
	int done;
	z_stream strm = {0};
	int ret;
	struct predictor pred = {1, 1, 8, 1};
	int (*depredict)(struct predictor *, uint8_t *, size_t);

	/* set up the predictor (if any) */
	#define SETPARM(VAR,STR) do {					\
		v = dictentry(parms, (STR));				\
		if (v != NULL) {					\
			if (v->token_type != TT_SINT || v->sint < 0)	\
				return NULL;				\
			VAR = v->sint;					\
		} } while(0)
	SETPARM(pred.num,	"Predictor");
	SETPARM(pred.colors,	"Colors");
	SETPARM(pred.bpc,	"BitsPerComponent");
	SETPARM(pred.columns,	"Columns");
	#undef SETPARM
	if (pred.num == 1)
		depredict = depred_none;
	else {
		if (pred.num >= 10 && pred.num <= 15)
			depredict = depred_png;
		else if (pred.num == 2) {
			/* for 8-bpc TIFF pred. 2, we can reuse PNG Sub */
			if (pred.bpc == 8) {
				pred.predfun = pp_sub;	/* predict left */
				depredict = depred_png;
			} else {
				// XXX add general TIFF predictor (bpc != 8)
				log_message(SEV_DONTCARE, "FlateDecode: /Predictor %d "
				    "not supported for /BitsPerComponent %d\n",
				    pred.num, pred.bpc);
				return NULL;
			}
		} else {
			log_message(SEV_DONTCARE, "FlateDecode: /Predictor %d"
			    " not supported\n", pred.num);
			return NULL;
		}

		if(pred.colors < 1)
		{
			log_message(9, "FlateDecode: /Colors has an invalid value of %d\n", pred.colors);
			return NULL;
		}
		if(pred.bpc != 1 && pred.bpc != 2 && pred.bpc != 4 && pred.bpc != 8 && pred.bpc != 16)
		{
			log_message(9, "FlateDecode: /BitsPerComponent has an invalid value of %d\n", pred.bpc);
			return NULL;
		}
		// XXX spec does not disallow nonpositive /Columns (7.4.4.3)
		// cf. https://github.com/pdf-association/pdf-issues/issues/15

		/* allocate row buffer */
		if (pred.columns > (INT_MAX - 7) / pred.colors / pred.bpc) {
			log_message(SEV_DONTCARE, "FlateDecode: overflow\n");
			return NULL;
		}
		pred.rowsz = (pred.colors * pred.bpc * pred.columns + 7) / 8;
		pred.buf = calloc(1, pred.rowsz);
		if (pred.buf == NULL)
			err(2, "FlateDecode");
	}

	/* set up zlib */
	// XXX pass our allocator to zlib
	ret = inflateInit(&strm);
	if (ret != Z_OK) {
		/* provide an error message if zlib didn't */
		if (ret != Z_OK && strm.msg == NULL) {
			switch (ret) {
			case Z_MEM_ERROR:
				strm.msg = "not enough memory";
				break;
			case Z_STREAM_ERROR:
				strm.msg = "invalid parameters";
				break;
			case Z_VERSION_ERROR:
				strm.msg = "library version mismatch";
				break;
			default:
				strm.msg = "unrecognized error";
				break;
			}
		}
		errx(2, "inflateInit: %s (returned %d)", strm.msg, ret);
	}
	buf = malloc(BUFSIZE);
	if (buf == NULL)
		err(2, "FlateDecode");

#ifdef ITERATIVE // XXX
	/* initialize target parser */
	sp = h_parse_start(p);
	assert(sp != NULL);
	pred.sp = sp;
#endif

	done = 0;
	strm.avail_in = b.len;
	strm.next_in = (unsigned char *)b.token;
	do {
		strm.avail_out = BUFSIZE;
		strm.next_out = buf;

		ret = inflate(&strm, Z_NO_FLUSH);
		if (ret != Z_STREAM_END && ret != Z_OK) {
			log_message(9, "inflate: %s (%d)\n", strm.msg, ret);
			break;
		}

		sz = BUFSIZE - strm.avail_out;
		done = depredict(&pred, buf, sz);
	} while (done == 0 && ret == Z_OK);

#ifdef ITERATIVE // XXX
	res = h_parse_finish(sp);
		// XXX always return NULL on error?
#else

	res = h_parse(p, pred.out, pred.nout);
	if (!res) {
		log_message(9, "\n\nFlateDecode:: Failed to parse stream of length %lu\n", pred.nout);
	}
	free(pred.out);
#endif
	inflateEnd(&strm);
	free(pred.buf);
	free(buf);

	if (done == -1)
		return NULL;
	return res;
}


#include "lzw.h"

// XXX refactor this copy-pastry (of FlateDecode)
HParseResult *
LZWDecode(const Dict *parms, HBytes b, HParser *p)
{
	struct predictor pred = {1, 1, 8, 1};
	int (*depredict)(struct predictor *, uint8_t *, size_t);
	HParseResult *res;
	HParseResult *tmp_res;
	int done;
	//int ret;
	const HParsedToken *v;
	int earlychange;

	/* set up the predictor (if any) */
	#define SETPARM(VAR,STR) do {					\
		v = dictentry(parms, (STR));				\
		if (v != NULL) {					\
			if (v->token_type != TT_SINT || v->sint < 0)	\
				return NULL;				\
			VAR = v->sint;					\
		} } while(0)
	SETPARM(pred.num,	"Predictor");
	SETPARM(pred.colors,	"Colors");
	SETPARM(pred.bpc,	"BitsPerComponent");
	SETPARM(pred.columns,	"Columns");
	#undef SETPARM
	if (pred.num == 1)
		depredict = depred_none;
	else {
		if (pred.num >= 10 && pred.num <= 15)
			depredict = depred_png;
		else if (pred.num == 2) {
			/* for 8-bpc TIFF pred. 2, we can reuse PNG Sub */
			if (pred.bpc == 8) {
				pred.predfun = pp_sub;	/* predict left */
				depredict = depred_png;
			} else {
				// XXX add general TIFF predictor (bpc != 8)
				log_message(SEV_DONTCARE, "LZWDecode: /Predictor %d "
				    "not supported for /BitsPerComponent %d\n",
				    pred.num, pred.bpc);
				return NULL;
			}
		} else {
			log_message(SEV_DONTCARE, "LZWDecode: /Predictor %d"
			    " not supported\n", pred.num);
			return NULL;
		}

		if(pred.colors < 1)
		{
			log_message(9, "LZWDecode: /Colors has an invalid value of %d\n", pred.colors);
			return NULL;
		}
		if(pred.bpc != 1 && pred.bpc != 2 &&  pred.bpc != 4 && pred.bpc != 8 && pred.bpc != 16)
		{
			log_message(9, "LZWDecode: /BitsPerComponent has an invalid value of %d\n", pred.bpc);
			return NULL;
		}
		// XXX spec does not disallow nonpositive /Columns (7.4.4.3)

		/* allocate row buffer */
		if (pred.columns > (INT_MAX - 7) / pred.colors / pred.bpc) {
			log_message(SEV_DONTCARE, "LZWDecode: overflow\n");
			return NULL;
		}
		pred.rowsz = (pred.colors * pred.bpc * pred.columns + 7) / 8;
		pred.buf = calloc(1, pred.rowsz);
		if (pred.buf == NULL)
			err(2, "LZWDecode");
	}

	v = dictentry(parms, "EarlyChange");
	if(v != NULL && v->token_type == TT_SINT && v->sint == 0)
	{
		earlychange = 0;
	}
	else
	{
		earlychange = 1;
	}

	init_LZW_context(earlychange);
	tmp_res = parse_LZW_data(b.token, b.len);

	if(!tmp_res)
	{
		log_message(9, "parse error in LZWDecode filter\n");
		return NULL;
	}

	assert(tmp_res->ast->token_type == TT_BYTES);

	uint8_t * tmp_buf = malloc(sizeof(uint8_t) * tmp_res->ast->bytes.len);
	if (tmp_buf == NULL)
		err(2, "LZWDecode");
	memcpy(tmp_buf, tmp_res->ast->bytes.token, tmp_res->ast->bytes.len);
	done = depredict(&pred, tmp_buf, tmp_res->ast->bytes.len);
	assert(!done);

	//done = depredict(&pred, res->ast->bytes.token, res->ast->bytes.len);
	//assert(!done);

	// SR::TODO:: Do a H_MAKE rather than a parse and let the caller do the parse
	res = h_parse(p, pred.out, pred.nout); // XXX: should kstream try to decode streams with no Type?
	//res = h_parse(p, tmp_res->ast->bytes.token, tmp_res->ast->bytes.len); // XXX depred buffer
	free(pred.out);
	free(tmp_buf);

	return res;
}

HParseResult *
RunLengthDecode(const Dict *parms, HBytes b, HParser *p)
{
	HParseResult *res;

	res = h_parse(p_rldstring, b.token, b.len);
	if(!res)
	{
		log_message(9, "parse error in RunLengthDecode filter\n");
		return NULL;
	}

	assert(res->ast && res->ast->token_type == TT_BYTES);
	// SR::TODO:: Do a H_MAKE rather than a parse and let the caller do the parse
	res = h_parse(p, res->ast->bytes.token, res->ast->bytes.len);

	return res;
}

/*
 * Decodes ASCII hexadecimal data into binary data.
 * parms should be empty, because the filter has no parameters
 */
HParseResult *
ASCIIHexDecode(const Dict *parms, HBytes b, HParser *p)
{
	HParseResult *res;

	res = h_parse(p_ahexstream, b.token, b.len);
	if(!res) {
		log_message(9, "parse error in ASCIIHexDecode filter\n");
		return NULL;
	}

	assert(res->ast && res->ast->token_type == TT_BYTES);
	// SR::TODO:: Do a H_MAKE rather than a parse and let the caller do the parse
	res = h_parse(p, res->ast->bytes.token, res->ast->bytes.len);

	return res;
}

/*
 * Decodes ASCII base-85 encoded data and produces binary data.
 * parms should be empty, because the filter has no parameters
 */
HParseResult*
ASCII85Decode(const Dict *parms, HBytes b, HParser *p)
{
	HParseResult *res;

	res = h_parse(p_ascii85, b.token, b.len);
	if(!res) {
		log_message(9, "parse error in ASCII85Decode filter\n");
		return NULL;
	}

	assert(res->ast && res->ast->token_type == TT_BYTES);
	// SR::TODO:: Do a H_MAKE rather than a parse and let the caller do the parse
	res = h_parse(p, res->ast->bytes.token, res->ast->bytes.len);

	return res;
}

HParsedToken *
act_rest(const HParseResult *p, void *env)
{
	struct Env *aux = env;
	size_t offset = H_CAST_UINT(p->ast) / 8;

	return H_MAKE_BYTES(aux->input + offset, aux->sz - offset);
}

HParser *
p_rest__m(HAllocator *mm__, struct Env *aux)
{
	return h_action__m(mm__, h_tell__m(mm__), act_rest, aux);
}

/* combine current position with env=(input,sz) into HBytes */
HParsedToken *
act_take_bytes(const HParseResult *p, void *env)
{
	const HBytes *bs = env;
	size_t offset = H_CAST_UINT(p->ast) / 8;

	/*
	 * NB: we must allocate a new HBytes struct here because the old one is
	 * allocated only temporarily for the lifetime of the continuation
	 * below.
	 */
	return H_MAKE_BYTES(bs->token + offset, bs->len);
}

HParser *
p_take__m(HAllocator *mm__, size_t n, struct Env *aux)
{
	HParser *skip, *bytes;
	HBytes *bs;

	/* dummy struct to hold the pair (input,n) */
	bs = h_alloc(mm__, sizeof(HBytes));
	bs->token = aux->input;
	bs->len = n;

	bytes = h_action__m(mm__, h_tell__m(mm__), act_take_bytes, bs);
	skip  = h_skip__m(mm__, n * 8);

	return h_left__m(mm__, bytes, skip);
}


// Parser for object streams
HParser *p_objstm__m(HAllocator *, const Dict *);

// Action for stream continuation
HParsedToken *act_ks_value(const HParseResult *p, void *u);
struct streamspec {
	Dict *dict;		/* stream dictionary */
	HParser *parser;	/* data parser */
};



/*
 * ********************************************************************
 * Start Catalog parsing
 * ********************************************************************
 */

bool
is_parent(Dict *dict, const HParsedToken *expected)
{
	const HParsedToken *item;
	Ref *ref, *pRef;
	bool res=false;

	item = dictentry(dict, "Parent");
	if (item == NULL) {
		res = item == expected;
	}
	else if (item->token_type == TT_Ref) {
		ref = H_CAST(Ref, item);
		if (expected == NULL) {
			return res;
		}
		else {
			assert(expected->token_type == TT_Ref);
			pRef = (Ref *)expected->user;
			res = (ref->nr == pRef->nr) && (ref->gen == pRef->gen);
		}
	}

	return res;
}

bool
has_value(Dict *dict, char *fn, char *value)
{
	const HParsedToken *item;
	bool res=false;

	item = dictentry(dict, fn);
	if ( (item != NULL) && (item->token_type == TT_BYTES) &&
			(bytes_eq(item->bytes, value)) ) {
		res = true;
	}

	return res;
}



Dict *
get_fontdict(const HParsedToken *obj, struct Env* aux)
{
	const HParsedToken *item;
	Ref                *ref;
	Dict *dict = NULL;
	ObjStm *stm = NULL;


	assert(obj->token_type == TT_Ref);
	ref = H_CAST(Ref, obj);
	item = resolve(aux, obj);
	if ( (item)  && (item->token_type == TT_Dict) ) {
		dict = H_CAST(Dict, item);
		if (! has_value(dict, "Type", "Font"))
			dict = NULL;
	}
	else if ( (item)  && (item->token_type == TT_ObjStm) ) {
		stm = H_CAST(ObjStm, item);
		for (int i=0; i<stm->numObjs; i++) {
			if ( (stm->tok[i].oid.nr == ref->nr) &&
					(stm->tok[i].oid.gen == ref->gen) ) {
				if (stm->tok[i].obj->token_type != TT_Dict) {
					dict = NULL;
				}
				else {
					dict = H_CAST(Dict, stm->tok[i].obj);
					if (! has_value(dict, "Type", "Font"))
						dict = NULL;
				}
				break;
			}
		}
	}


	return dict;
}

const HParsedToken *
get_dictoftype(
		const HParsedToken *obj,
		const HParsedToken *parent_t,
		char               *value,
		struct Env         *aux)
{
	Dict *dict = NULL;
	const HParsedToken *tok;
	ObjStm *stm = NULL;

	if (obj->token_type == TT_Dict) {
		dict = H_CAST(Dict, obj);
		if (is_parent(dict, parent_t) && has_value(dict, "Type", value))
			tok = obj;
		else
			dict = NULL;
	}
	else if (obj->token_type == TT_ObjStm) {
		stm = H_CAST(ObjStm, obj);
		for (int i=0; i<stm->numObjs; i++) {
			const HParsedToken *sitem = resolve(aux, stm->tok[i].obj);
			if ((sitem) && (sitem->token_type == TT_Dict)) {
				dict = H_CAST(Dict, sitem);
				if (is_parent(dict, parent_t) && has_value(dict, "Type", value)) {
					tok = sitem;
					break;
				}
				else
					dict = NULL;
			}
		}
	}

	if (dict == NULL)
		tok = NULL;

	return tok;
}





Fontinfo_T *
getFontinfo(const Dict *fontdict, char *name, struct Env *aux)
{

	assert (fontdict);
	Fontinfo_T *fontinfo = h_arena_malloc(fontdict->arena, sizeof(Fontinfo_T));
	fontinfo->name            = NULL;
	fontinfo->type            = NULL;
	fontinfo->basefont        = NULL;
	fontinfo->encoding        = NULL;
	fontinfo->descriptor      = NULL;
	fontinfo->toUnicode       = NULL;
	fontinfo->descendantFonts = NULL;

	assert(fontinfo);
	const HParsedToken *item = dictentry(fontdict, "Name");
	if (item) {
		assert(item->token_type == TT_BYTES);
		if (bytes_eq(item->bytes, name)) {
			fontinfo->name = h_arena_malloc(fontdict->arena, sizeof(char)*(item->bytes.len+1));
			memcpy(fontinfo->name, (char *)item->bytes.token, item->bytes.len);
			fontinfo->name[item->bytes.len] = '\0';
		}
	}
	item = dictentry(fontdict, "Subtype");
	if (item) {
		assert (item->token_type == TT_BYTES);
		fontinfo->type = h_arena_malloc(fontdict->arena, sizeof(char)*(item->bytes.len+1));
		memcpy(fontinfo->type, (char *)item->bytes.token, item->bytes.len);
		fontinfo->type[item->bytes.len] = '\0';
	}
	item = dictentry(fontdict, "BaseFont");
	if (item) {
		assert (item->token_type == TT_BYTES);
		fontinfo->basefont = h_arena_malloc(fontdict->arena, sizeof(char)*(item->bytes.len+1));
		memcpy(fontinfo->basefont, (char *)item->bytes.token, item->bytes.len);
		fontinfo->basefont[item->bytes.len] = '\0';
	}
	item = dictentry(fontdict, "Encoding");
	item = resolve(aux, item);
	if (item) {  // TODO: Failure ==> xref error -- Figure out how to handle
		if (item->token_type == TT_BYTES) {
			fontinfo->encoding = h_arena_malloc(fontdict->arena, sizeof(char)*(item->bytes.len+1));
			memcpy(fontinfo->encoding, (char *)item->bytes.token, item->bytes.len);
			fontinfo->encoding[item->bytes.len] = '\0';
		}
		else if (item->token_type == TT_Dict)
		{
			const Dict *encodingDict = H_CAST(Dict, item);
			item = dictentry(encodingDict, "BaseEncoding");
			if (item) {
				fontinfo->encoding = h_arena_malloc(fontdict->arena, sizeof(char)*(item->bytes.len+1));
				memcpy(fontinfo->encoding, (char *)item->bytes.token, item->bytes.len);
				fontinfo->encoding[item->bytes.len] = '\0';
			}
		}
		else {
			//fprintf(stdout, "\nUnexpected token type in parsing font -Encoding- attribute -"
			//		"token_type = %u\n", item->token_type);
		}
	}

	item = dictentry(fontdict, "FontDescriptor");
	item = resolve(aux, item);
	if (item) { // TODO: Failure ==> xref error -- Figure out how to handle
		item = get_dictoftype(item, NULL, "FontDescriptor", aux);
		if (item)
			fontinfo->descriptor = item;
	}

	item = dictentry(fontdict, "ToUnicode");
	item = resolve(aux, item);
	if (item) { // TODO: Failure ==> xref error -- Figure out how to handle
		item = get_dictoftype(item, NULL, "ToUnicode", aux);
		if (item)
			fontinfo->toUnicode = item;
	}

	item = dictentry(fontdict, "DescendantFonts");
	item = resolve(aux, item);
	if (item) { // TODO: Failure ==> xref error -- Figure out how to handle
		item = get_dictoftype(item, NULL, "DescendantFonts", aux);
		if (item)
			fontinfo->descendantFonts = item;
	}

	return fontinfo;
}

Fontinfo_T *
lookup_font(TextState_T *state, struct Env *aux)
{
	const HParsedToken *item = NULL;
	Dict               *fontlist; // font list dictionary in page
	Fontinfo_T         *fontinfo = NULL;

	PtNode_T *page   = state->node;
	assert ( (page->type == PG_NODE) || (page->type == XO_NODE) );
	struct TextEntry_S *fentry = state->font;
	if (fentry) {
		assert ( fentry->type == TS_Tf);
		struct fontref *fr = &fentry->fref;
		char      *fn = fr->fn;
		Dict       *dict = NULL;
		if (page->pgRsrc  && (page->pgRsrc->numFonts > 0)) {
			fontlist = H_CAST(Dict, page->pgRsrc->fonts);
			assert(page->pgRsrc->numFonts == fontlist->used);
			for (int i=0; i< page->pgRsrc->numFonts; i++) {
				item = dictentry(fontlist, fn);  // look for the font name in the dictionary */
				if (item) {
					dict = get_fontdict(item, aux);
					if (dict) {
						fontinfo = getFontinfo(dict, fr->fn, aux);
					}
				}

			}
		}
	}
	return (fontinfo);
}

bool parse_fonts(const HParsedToken *dict_t, RsrcDict_T *pgRsrc, struct Env *aux)
{
	Dict               *fontdict = H_CAST(Dict, dict_t);
	const HParsedToken *item = NULL;
	bool                processed = false;

	// Handle a font resource
	if (has_value(fontdict, "Type", "Font")) {
		if (pgRsrc->fonts) {
			log_message(SEV_DONTCARE, "\n\nparse_fonts: Attempt to add fonts -- Supported??\n\n");
		}
		else {
			pgRsrc->fonts = item;
			pgRsrc->numFonts += 1;
		}
		return true;
	}



	// Handle a dictionary of font resources
	item = dictentry(fontdict, "Font");
	item = resolve(aux, item);
	if (item) { // TODO: Failure ==> xref error -- Figure out how to handle
		assert(item->token_type == TT_Dict);
		processed = true;
		Dict *fontlist = H_CAST(Dict, item);
		if (pgRsrc->fonts) {
			log_message(SEV_DONTCARE, "\n\nparse_fonts: Attempt to add fonts -- Supported??\n\n");
		}
		else {
			pgRsrc->fonts = item;
			pgRsrc->numFonts = fontlist->used;
		}
	}
	return processed;
}

RsrcDict_T *parse_rsrcdict(HArena *arena, const HParsedToken *dict_t, PtNode_T *pgNode, struct Env *aux);


XoNode_T*
create_XoNode(HArena *arena, Catalog_T *catalog)
{
	XoNode_T *node = h_arena_malloc(arena, sizeof(XoNode_T));
	node->name = NULL;
	node->node = NULL;
	node->next = NULL;
	if (catalog->xoHead == NULL)
		catalog->xoHead = node;
	if (catalog->xoTail == NULL)
		catalog->xoTail = node;
	else {
		catalog->xoTail->next = node;
		catalog->xoTail       = node;
	}
	catalog->xoCount += 1;

	return node;
}


// XObject resources can be recursively specified
void parse_xobject(
		const HParsedToken *dict_t,
		PtNode_T           *parent,
		RsrcDict_T         *pgRsrc,
		struct Env         *aux)
{

	Dict               *xodict = H_CAST(Dict, dict_t);
	const HParsedToken *item   = NULL;   // generic token
	const HParsedToken *tok    = NULL;   // resolved token
	const Dict         *xobj_d = NULL;  // dictionary associated with reference token
	const HParsedToken *xobj_t = NULL;  // xobject resource token


	item = dictentry(xodict, "XObject");
	item = resolve(aux, item);
	if (item) {  // test for no XObject key (TODO: Failure ==> xref error)
		assert(item->token_type == TT_Dict);

		bool isaFontRsrc = parse_fonts(item, pgRsrc, aux);
		if (isaFontRsrc)
			return;

		Dict *xolist = H_CAST(Dict, item);

		aux->catalog.xoCount += xolist->used;
		// work on each element of the dictionary
		for (int i=0; i<xolist->used; i++) {
			XoNode_T *xobj_r = create_XoNode(xodict->arena, &aux->catalog);
			HBytes k = H_INDEX_BYTES(xolist->elements[i], 0);
			xobj_r->name = h_arena_malloc(xodict->arena, k.len+1);

			memcpy(xobj_r->name, k.token, k.len);
			xobj_r->name[k.len] = '\0';
			const HParsedToken *ref = H_INDEX_TOKEN(xolist->elements[i], 1);
			assert(ref->token_type == TT_Ref);
			xobj_r->node =  h_arena_malloc(xodict->arena, sizeof(PtNode_T));
			xobj_r->node->type     = XO_NODE;
			xobj_r->node->parent_n = parent;
			xobj_r->node->parent_t = parent->me_t;
			xobj_r->node->pgRsrc   = pgRsrc;
			xobj_r->node->me_t     = ref;
			tok = resolve(aux, ref);

			if (tok == NULL) continue;

			// tok can be an image dictionary -- which we are ignoring
			if (tok->token_type == TT_SEQUENCE)
				xobj_t = H_INDEX_TOKEN(tok, 0);  // expecting an XObject dictionary token
			if (xobj_t == NULL) continue;

			xobj_t = get_dictoftype(xobj_t, NULL, "XObject", aux); // test it
			if (xobj_t == NULL) continue;

			xobj_d = H_CAST(Dict, xobj_t);
			item = dictentry(xobj_d, "Subtype");
			if (item == NULL || item->token_type != TT_BYTES)
				continue; // no "Subtype" field

			/*
			 *  TODO:: external objects can be images, forms, or postscript objects
			 *  We are only handling forms at the moment
			 */
			if (bytes_eq(item->bytes, "Form")) {
				const HParsedToken *rsrcdict;

				rsrcdict = dictentry(xobj_d, "Resources");
				rsrcdict = resolve(aux, rsrcdict);
				if (!rsrcdict) {
					xobj_r->node->xn.textStream = NULL;
					continue;
				}

				xobj_r->node->pgRsrc = parse_rsrcdict(xodict->arena, rsrcdict, xobj_r->node, aux);

				// set the text state to this xobject
				// parse the text stream, which is field 2 of the sequence
				aux->curr_node = xobj_r->node;
				xobj_t = H_INDEX_TOKEN(tok, 1);  // expecting an HParseResult token
				const HParseResult *res = H_CAST(HParseResult, xobj_t);

				/* Text stream is missing from object or failed to parse */
				if (!res)
				{
					xobj_r->node->xn.textStream = NULL;
					continue;
				}
				else
				{
					HBytes stm = H_CAST_BYTES(res->ast);
					res = h_parse(p_textstream, stm.token, stm.len);

					if (res) { // text found in stream
//							xobj_r->node->xn.dict       = 0;
						xobj_r->node->xn.textStream = res->ast;
					}
				}
			}
		}
	}
}


RsrcDict_T *parse_rsrcdict(HArena *arena, const HParsedToken *dict_t, PtNode_T *pgNode, struct Env *aux)
{
	RsrcDict_T *rsrc = NULL;


	// Process the dictionary
	if ( (dict_t->token_type == TT_Dict) || (dict_t->token_type == TT_ObjStm) ) {
		rsrc = h_arena_malloc(arena, sizeof(RsrcDict_T));
		rsrc->resources   = dict_t;
		rsrc->fonts       = NULL;
		rsrc->numFonts    = 0;
		rsrc->xobj        = NULL;
//		rsrc->seenCmaps   = NULL;
//		rsrc->numCmapsSeen = 0;
	}

	// Resource is a simple dictionary
	if (dict_t->token_type == TT_Dict) {
		parse_fonts(dict_t, rsrc, aux);
		parse_xobject(dict_t, pgNode, rsrc, aux);
	}
	else if (dict_t->token_type == TT_ObjStm) {
		const ObjStm *strmc = H_CAST(ObjStm, dict_t);

		for (int i=0; i<strmc->numObjs; i++) {
			if (strmc->tok[i].obj->token_type == TT_Dict) {
				parse_fonts(strmc->tok[i].obj, rsrc, aux);
				parse_xobject(strmc->tok[i].obj, pgNode, rsrc, aux);
			}
		}
	}

	return (rsrc);

}


void pp_pgrsrc(FILE *stream, const RsrcDict_T *pgRsrc)
{
	if (pgRsrc) {
		fprintf(stream, "\npp_pgrsrc: Num fonts used in this page = %lu\n", pgRsrc->numFonts);
		fprintf(stream, "pp_pgrsrc: Resources\n");
		if (pgRsrc->resources) h_pprintln(stream, pgRsrc->resources);
		fprintf(stream, "pp_pgrsrc: Fonts\n");
		if (pgRsrc->fonts)     h_pprintln(stream, pgRsrc->fonts);
		fprintf(stream, "pp_pgrsrc: XObjects\n");
		if (pgRsrc->xobj)      h_pprintln(stream, pgRsrc->xobj);
	}
}

void pp_ptnode(FILE *stream, const PtNode_T *node)
{
	fprintf(stream, "\nPage Tree Node Info:\n");
	fprintf(stream, "pp_ptnode: parent = ");
	fprintf(stream, "pp_ptnode: parent_ref = %p\n", node->parent_n);
	if (node->parent_t) h_pprintln(stream, node->parent_t);
	fprintf(stream, "\npp_ptnode: me_ref = ");
	if (node->me_t)     h_pprintln(stream, node->me_t);
	if (node->pgRsrc) pp_pgrsrc(stream, node->pgRsrc);
}

void parse_pagenode (
		struct Env         *aux,
		PtNode_T           *myNode,
		const HParsedToken *myRef,     // my page node reference
		const Dict         *myDict,    // my page node specification
		const HParsedToken *parent_t,  // parent reference token
		PtNode_T           *parent_n,  // my parent node
		HArena             *arena
		)
{

	const HParsedToken *contents    = NULL;


    // Initialize the page tree node
    myNode->type               = PG_NODE;
    myNode->parent_n           = parent_n;  // NULL for root
    myNode->parent_t           = parent_t;
    myNode->me_t               = myRef;
    aux->catalog.lastNode++;  // keep track of the node count
    aux->catalog.lastPage++;  // keep track of the page count
    aux->curr_node            = myNode;
    myNode->ts.node           = myNode;  // set the current page to me
    myNode->ts.font           = NULL;
    myNode->ts.char_spacing   = 0.0;
    myNode->ts.word_spacing   = 0.0;
    myNode->ts.line_spacing   = 0.0;
    // locate the starting point at top left of a 8.5x11" paper
    myNode->ts.curr_pos.tx    = 0.0;   // 8.5" = 612 points
    myNode->ts.curr_pos.ty    = 0.0; // 11"  = 792 points
    myNode->pn.dict            = myDict;
    myNode->nodeNum            = aux->catalog.lastNode;
    myNode->pn.page_num        = aux->catalog.lastPage;
    myNode->pn.textStream      = NULL;
    // initialize by inheriting -- override with local
    myNode->pgRsrc             = parent_n->pgRsrc;
    myNode->mediaBox.tx        = parent_n->mediaBox.tx;
    myNode->mediaBox.ty        = parent_n->mediaBox.ty;



    const HParsedToken *item        = NULL;
    item = dictentry(myDict, "Parent");
    bool matched = false;
    if (item && parent_t) { // neither item nor parent_t should be NULL
       assert(item->token_type == TT_Ref);
       if ( ( ((Ref*)item->user)->nr == ((Ref*)parent_t->user)->nr ) &&
             ( ((Ref*)item->user)->gen == ((Ref*)parent_t->user)->gen ) ) {
            matched = true;
        }
    }
    if (! matched) {
        fprintf(stdout, "parse_pagenode: Inconsistent or corrupt parent key!\n");
        fprintf(stdout, "Parent (from caller) =\n");
        h_pprintln(stdout, parent_t);
        fprintf(stdout, "Parent (from node) =\n");
        h_pprintln(stdout, item);

        // just continue for now
        return;
    }



	// Hold on to the mediaBox (page dimensions)-- required
	item = dictentry(myDict, "MediaBox"); // inheritable if not NULL
	item = resolve(aux, item);
	if (item && item->token_type == TT_SEQUENCE) {
		double llx, lly, urx, ury; // typical but can be any diagonal
		HParsedToken * token;
		token = H_INDEX_TOKEN(item, 0);
		if (token->token_type == TT_SINT)
			llx = token->sint;
		else
			llx = H_CAST_DOUBLE(token);
		token = H_INDEX_TOKEN(item, 1);
		if (token->token_type == TT_SINT)
			lly = token->sint;
		else
			lly = H_CAST_DOUBLE(token);
		token = H_INDEX_TOKEN(item, 2);
		if (token->token_type == TT_SINT)
			urx = token->sint;
		else
			urx = H_CAST_DOUBLE(token);
		token = H_INDEX_TOKEN(item, 3);
		if (token->token_type == TT_SINT)
			ury = token->sint;
		else
			ury = H_CAST_DOUBLE(token);

		myNode->mediaBox.tx = fabs(llx - urx);
		myNode->mediaBox.ty = fabs(lly - ury);
	}

	// Hold on to the Resources dictionary -- required
	const HParsedToken *rsrcDictT;
	item = dictentry(myDict, "Resources");
	rsrcDictT = resolve(aux, item);
	if (rsrcDictT)
		myNode->pgRsrc = parse_rsrcdict(myDict->arena, rsrcDictT, myNode, aux);




	// Process the contents stream or array  -- optional
	contents = dictentry(myDict, "Contents");
	//contents = resolve(aux, contents);
		// XXX reinstate when we can distinguish stream objects from
		//     arrays (TT_SEQUENCE). code below currently only handles
		//     the case where /Contents is a direct object that happens
		//     to be an array.
	if (contents == NULL)
		return;
	// XXX handle single stream case
	if (contents->token_type == TT_SEQUENCE) {	/* array case */
		size_t   numelts = contents->seq->used;
		size_t   bufsz   = 0;
		HBytes   bstrm;
		const HParsedToken *entry = NULL;
		const HParsedToken **pieces;

		pieces = h_arena_malloc(arena, sizeof(HBytes*) * numelts);
		for (int i=0; i<numelts; i++) {
			entry = H_INDEX_TOKEN(contents, i);
			entry = resolve(aux, entry);
			// XXX check that entry is a stream object of the expected type
			if (!entry) // TODO: Failure ==> xref error -- Figure out how to handle
				return;
			HParsedToken *res_strm = H_INDEX_TOKEN(entry, 1);
			HParseResult *bstrm_r = H_CAST(HParseResult, res_strm);
			if ( (bstrm_r) && (bstrm_r->ast) ) {
				pieces[i] = bstrm_r->ast;
				bstrm = H_CAST_BYTES(bstrm_r->ast);
				bufsz += bstrm.len;
			}
			else {
				const HParsedToken *nullbyte = h_make_bytes(arena, NULL, 0);
				pieces[i] = nullbyte;
			}

		}
		uint8_t *whole = h_arena_malloc(arena, sizeof(uint8_t) * bufsz);
		size_t offset=0;
		for (int i=0; i<numelts; i++) {
			bstrm = H_CAST_BYTES(pieces[i]);

			if (bstrm.len) {
				memcpy(&whole[offset], bstrm.token, bstrm.len);
				offset+=bstrm.len;
			}
		}
		assert(offset == bufsz);
		HParseResult *tstrm=h_parse(p_textstream, whole, bufsz);
		if (tstrm)
			myNode->pn.textStream = tstrm->ast;
	}
}



/*
 * This helper implements the standard backwards parsing strategy to read
 * the trailer dictionaries found at the very end of the input.
 *
 * It then follows the catalog dictionary to enumerate the pages in the pdf file
 * identifying text streams and contents streams, saving the information to support
 * text extraction in the environment structure.
 *
 * Parent reference and parent node are both NULL for the page tree
 *
 * A return value of false indicates some parsing error.
*/
void
parse_pagetree(
		struct Env         *aux,
		PtNode_T           *myNode,    // my page tree node
		const HParsedToken *myRef,     // my page tree node reference
		const Dict         *myDict,    // my page tree specification
		const HParsedToken *parent_t,     // my parent reference token
		PtNode_T           *parent_n    // my parent node
		)
{
	// Initialize the page tree node
    myNode->type     = PG_TREE;
    myNode->parent_n = parent_n;  // NULL for root
    myNode->parent_t = parent_t;
    myNode->me_t     = myRef;
    aux->catalog.lastNode++;  // keep track of the node count
    myNode->nodeNum  = aux->catalog.lastNode;
    if (parent_n) { // inheritable if in a tree node
        myNode->pgRsrc      = parent_n->pgRsrc;
        myNode->mediaBox.tx = parent_n->mediaBox.tx;
        myNode->mediaBox.ty = parent_n->mediaBox.ty;
    }
    else {
        myNode->pgRsrc      = NULL;
        myNode->mediaBox.tx = 0;
        myNode->mediaBox.ty = 0;
    }


    // make sure that the parent node matches
    const HParsedToken *item       = NULL;
	item = dictentry(myDict, "Parent");  // if root node ==> parent should be NULL
	bool matched = false;
	if (item && parent_t) {
        if ( ( ((Ref*)item->user)->nr == ((Ref*)parent_t->user)->nr ) &&
             ( ((Ref*)item->user)->gen == ((Ref*)parent_t->user)->gen ) ) {
            matched = true;
        }
	}
	else {   // either the parent is the page tree root or the node does not have a valid Parent key
	    if (item == parent_t) {// if item is NULL, we expect parent_t to also be NULL (==> root)
	        matched = true;
	    }
	}
	if (! matched) {
        fprintf(stdout, "parse_pagetree: Inconsistent or corrupt parent key!\n");
        fprintf(stdout, "Parent (from caller) =\n");
        h_pprintln(stdout, parent_t);
        fprintf(stdout, "Parent (from node) =\n");
        h_pprintln(stdout, item);

        // just continue for now
        return;
	}


	// Count is a required field
	item = dictentry(myDict, "Count");
	item = resolve(aux, item);  // anything can be a ref
	if (item == NULL || item->token_type != TT_SINT) {
		fprintf(stdout, "parse_pagetree: Required page node count "
		    "missing or corrupt!\n");
		return; // Should just be a warning?
	}
	myNode->pt.numPages = H_CAST_SINT(item);



	// page tree node may have resources and media box that are inheritable
	item = dictentry(myDict, "MediaBox"); // inheritable if not NULL
	item = resolve(aux, item);
	if (item && item->token_type == TT_SEQUENCE) {
		double llx, lly, urx, ury; // typical but can be any diagonal
		HParsedToken * token;
		token = H_INDEX_TOKEN(item, 0);
		if (token->token_type == TT_SINT)
			llx = token->sint;
		else
			llx = H_CAST_DOUBLE(token);
		token = H_INDEX_TOKEN(item, 1);
		if (token->token_type == TT_SINT)
			lly = token->sint;
		else
			lly = H_CAST_DOUBLE(token);
		token = H_INDEX_TOKEN(item, 2);
		if (token->token_type == TT_SINT)
			urx = token->sint;
		else
			urx = H_CAST_DOUBLE(token);
		token = H_INDEX_TOKEN(item, 3);
		if (token->token_type == TT_SINT)
			ury = token->sint;
		else
			ury = H_CAST_DOUBLE(token);
		myNode->mediaBox.tx = fabs(llx - urx);
		myNode->mediaBox.ty = fabs(lly - ury);
	}

	const HParsedToken *rsrcDictT  = NULL;
	item = dictentry(myDict, "Resources"); // inheritable if not NULL
	rsrcDictT = resolve(aux, item);
	if (rsrcDictT)
		myNode->pgRsrc = parse_rsrcdict(myDict->arena, rsrcDictT, myNode, aux);





	// Kids array is a required field -- process the kids (pgTable)
	// It should be an array of references to page or page tree nodes.
	const HParsedToken *kids = NULL;
	kids = dictentry(myDict, "Kids");
	kids = resolve(aux, kids);
	if (kids == NULL || kids->token_type != TT_SEQUENCE) {
		fprintf(stdout, "parse_pagetree: Tree node has no pages!\n");
		return;
	}


	PtNode_T           *kid       = NULL;
	const HParsedToken *kidRef    = NULL; // page tree or page node ref.
	const HParsedToken *kidDict_t = NULL;
	const HParsedToken *pageDict_t = NULL;
	const HParsedToken *treeDict_t = NULL;
	const Dict         *kidDict   = NULL;

	HCountedArray  *pgTable = H_CAST_SEQ(kids);
	size_t          pgtSz   = pgTable->used;
	if (pgtSz != myNode->pt.numPages) {
	    fprintf(stdout, "parse_pagetree: Size mismatch:: "
	            "Kids array size = %lu, Page Count in node = %u\n",
	            pgtSz, myNode->pt.numPages);
	    // continue processing the pages
	}
	myNode->pt.kids  = (PtNode_T*)h_arena_malloc(pgTable->arena, pgtSz * sizeof(PtNode_T));

	// Process the kids
	for (int i=0; i<pgtSz; i++)
	{
		kid         = &myNode->pt.kids[i];
		kidRef      = pgTable->elements[i];
		kidDict_t   = resolve(aux, kidRef);     // page or tree node dictionary or object stream token

		if (kidDict_t) {
			// Look for a tree node
			treeDict_t = get_dictoftype(kidDict_t, myRef, "Pages", aux);
			if (treeDict_t) {
				kidDict = H_CAST(Dict, treeDict_t);
				parse_pagetree(aux, kid, kidRef, kidDict, myRef, myNode);
			}
			// Look for a page node
			pageDict_t = get_dictoftype(kidDict_t, myRef, "Page", aux);
			if (pageDict_t) {
				kidDict = H_CAST(Dict, pageDict_t);
				parse_pagenode(aux, kid, kidRef, kidDict, myRef, myNode, pgTable->arena);
			}
		}
	} // end loop


	return;
}




/*
 * This helper starts the process of elaborating the page tree
 * starting with the trailer dictionary
 */
void
parse_catalog(struct Env *aux, const HParsedToken *root)
{
	const HParsedToken *dict_t  = NULL;
	const Dict         *catalog = NULL;
	const HParsedToken *ptRef   = NULL; // page tree reference
	const Dict         *ptRoot  = NULL; // page tree root Dictionary


	// initialize the catalog structure
	aux->catalog.catalog    = NULL;
	aux->catalog.pRoot      = NULL;
	aux->catalog.lastPage  = 0;
    aux->catalog.lastNode  = 0;
    // Initialize the xobject structure
	aux->catalog.xObjs.name = NULL;
	aux->catalog.xObjs.node = NULL;
	aux->catalog.xObjs.next = NULL;
	aux->catalog.xoHead     = NULL;
	aux->catalog.xoTail     = NULL;
	aux->catalog.xoCount    = 0;


	// Ensure the reference is to the catalog dictionary
	dict_t = resolve(aux, root);
	if (!dict_t) { // TODO: Failure ==> xref error -- Figure out how to handle
		return;
	}

	aux->catalog.catalog = get_dictoftype(dict_t, NULL, "Catalog", aux); // catalog dictionary token
	if (aux->catalog.catalog) { // Caution:: relying on the short-circuiting behavior here
		catalog = H_CAST(Dict, aux->catalog.catalog);


		// Catalog found -- Now get the root of the page tree associated with the catalog
		ptRef = dictentry(catalog, "Pages"); // indirect reference to a dictionary
		if ( (ptRef == NULL) || (ptRef->token_type != TT_Ref) ) {
			return;
		}
		aux->catalog.pRoot = ptRef; // indirect reference to the page tree


		/* resolve and process the page tree root reference to extract the dictionary --> Page Tree Object */
		dict_t = resolve(aux, ptRef);  // page tree root node
		if (!dict_t) {
			return;
		}
		// make sure the retrieved token is a dictionary of /Type "Pages"
		dict_t = get_dictoftype(dict_t, NULL, "Pages", aux);   // page tree root dictionary (parent is NULL)
		if (!dict_t) {
			return; // Nothing more to do here
		}
		ptRoot = H_CAST(Dict, dict_t);

		// parse_pagetree
		parse_pagetree(aux, &aux->catalog.pgTree, ptRef, ptRoot, NULL, NULL);
	}
	else {   // looks like the field "Type:Catalog" is a hint, not a requirement for a valid pdf
		return;

	}
}

/*
 * ********************************************************************
 * End Catalog parsing
 * ********************************************************************
 */








/*
 * ********************************************************************
 * Start xref parsing
 * ********************************************************************
 */
/*
 * decode the bytes in 'b' according to metadata in the stream dictionary 'd'
 * and parse the result with 'p'.
 */
HParseResult *
decode_stream(const Dict *d, HBytes b, HParser *p)
{
	HParseResult *(*filter)(const Dict *, HBytes, HParser *);
	const Dict *parms = NULL;
	const HParsedToken *v;

	v = dictentry(d, "Filter");
	if (v == NULL)
		return h_parse(p, b.token, b.len);

#ifdef ITERATIVE // XXX
	/* compile to a CF backend to enable incremental parsing */
	if (h_compile(p, PB_LLk, NULL) == -1)
		errx(2, "stream data parser: LL(1) compile failed");
#endif

	if (v->token_type != TT_BYTES) {
		// XXX TT_SEQUENCE would be a filter chain; that’s not supported, yet.
		// But it might also be something bogus, in which case we should fail.
		return NULL;
	}

	if (bytes_eq(v->bytes, "FlateDecode"))
		filter = FlateDecode;
	else if (bytes_eq(v->bytes, "ASCIIHexDecode"))
		filter = ASCIIHexDecode;
	else if (bytes_eq(v->bytes, "ASCII85Decode"))
		filter = ASCII85Decode;
	else if (bytes_eq(v->bytes, "RunLengthDecode"))
		filter = RunLengthDecode;
	else if (bytes_eq(v->bytes, "LZWDecode"))
		filter = LZWDecode;
	else {		/* filter not supported */
		log_message(SEV_DONTCARE, "decode_stream:: Unsupported Filter [%.*s]\n",
				(int)v->bytes.len, v->bytes.token);
		return NULL; /* Treat the stream as a byte array */
	}

	v = dictentry(d, "DecodeParms");
	if (v && v->token_type == TT_Dict)
		parms = v->user;

	return filter(parms, b, p);
}


HParser *p_xrefdata__m(HAllocator *, const Dict *);


HParser *
p_stream_data__m(HAllocator *mm__, const Dict *dict, struct Env *aux)
{
	const HParsedToken *v;

	v = dictentry(dict, "Type");
	v = resolve(aux, v);
	if (v == NULL || v->token_type != TT_BYTES)	// XXX -> custom type
		return NULL;				/* no /Type field */

	/* interpret known stream types */
	if (bytes_eq(v->bytes, "XRef"))
		return p_xrefdata__m(mm__, dict);

	if (bytes_eq(v->bytes, "ObjStm")) {
		return p_objstm__m(mm__, dict);
	}

	if (bytes_eq(v->bytes, "XObject")) {
		/*
		 *  external objects can be images, forms, or postscript objects
		 *  Forms and postscript objects can be handled as bytestreams
		 *  Additional XObject Forms processing will be handled during page traversal
		 *  Explicitly avoid parsing Image objects to improve speed -- send back NULL
		 */
		v = dictentry(dict, "Subtype");
		if (v == NULL) {
			log_message(7, "\nFailed to parse Subtype of XObject in dictionary");
			return (NULL);
		}
		if (bytes_eq(v->bytes, "Form")) {
			return p_bytes;
		}
#if 0
		if (bytes_eq(v->bytes, "Image")) {
			return p_fail;
		}
#endif
	}
	return NULL;					/* unrecognized type */
}

HParsedToken *
act_ks_value(const HParseResult *p, void *u)
{
	struct streamspec *spec = u;
	HBytes bytes = H_CAST_BYTES(p->ast);
	HParseResult *res;

	/* decode and parse the stream data */
	res = decode_stream(spec->dict, bytes, spec->parser);
	if (res == NULL) {
		HBytes b = {NULL, 0};
		const HParsedToken *v = dictentry(spec->dict, "Type");
		if (v != NULL && v->token_type == TT_BYTES)
			b = v->bytes;
		if (b.len > INT_MAX)
			b.len = INT_MAX;
		log_message(7, "parse error in stream (type %*s)\n",
		    (int)b.len, b.token);
		// XXX return the undecoded stream (p->ast)?
	}
	return H_MAKE(HParseResult, res);
}

/*
 * This continuation takes the stream dictionary (as first element of x) and
 * should return a parser that consumes exactly the bytes that make up the
 * stream data.
 *
 * x = (dict ...)
 */
HParser *
kstream(HAllocator *mm__, const HParsedToken *x, void *env)
{
	struct Env *aux = env;
	HParsedToken *dict_t = H_INDEX_TOKEN(x, 0);
	Dict *dict = H_CAST(Dict, dict_t);
	const HParsedToken *v = NULL;
	HParser *bytes_p, *dict_p, *value_p;
	struct streamspec *spec;
	size_t sz;

	/* look for the Length entry */
	v = dictentry(dict, "Length");
	v = resolve(aux, v);		/* resolve indirect references */
	if (v == NULL || v->token_type != TT_SINT || v->sint < 0)
		goto fail;
	sz = (size_t)v->sint;

	dict_p	= p_return__m(mm__, dict_t);
	bytes_p = p_take__m(mm__, sz, aux);

	spec = h_alloc(mm__, sizeof(struct streamspec));
	spec->dict = dict;
	spec->parser = p_stream_data__m(mm__, dict, aux);

	if (spec->parser == NULL)
		spec->parser = p_bytes;

	value_p = h_action__m(mm__, bytes_p, act_ks_value, spec);
		// XXX this should be h_bind, i think?! so failures in the
		// postordinate parser propagate...

	return h_sequence__m(mm__, dict_p, value_p, NULL);

fail:
	if (v == NULL)
		log_message(7, "stream /Length missing\n");
	else if (v -> token_type != TT_SINT)
		log_message(7, "stream /Length not an integer\n");
	else if (v < 0)
		log_message(7, "stream /Length negative\n");

	return p_fail;
}

HParser *
p_xrefsub__m(HAllocator *mm__, size_t base, size_t count, HParser *p_entry)
{
	HParser *ret_base, *ret_count, *p_header, *p_entries;

	ret_base  = p_return_uint__m(mm__, base);
	ret_count = p_return_uint__m(mm__, count);
	p_header  = h_sequence__m(mm__, ret_base, ret_count, NULL);
	p_entries = h_repeat_n__m(mm__, p_entry, count);

	return h_sequence__m(mm__, p_header, p_entries, NULL);
}

HParser *
p_xrefdata__m(HAllocator *mm__, const Dict *dict)
{
	const HParsedToken *v;
	HParser *p_field[3], *p_entry, **p_subs;
	size_t W[3];
	size_t Size, Wn, Wskip;

	/*
	 * what follows is a horrible bunch of code that builds, from the
	 * entries /W, /Index, and /Size in the stream dictionary, a parser for
	 * the cross-reference data itself.
	 *
	 * in short, every cross-reference entry consists of (as of PDF 2.0)
	 * three fields, but it could be more. /W gives the widths (in bytes)
	 * of these fields. the /Index specifies the division of the data into
	 * subsections; it is an array of natural numbers that in pairs specify
	 * the base object number and length of each subsection - analogous to
	 * the subsection headers in classic xref sections.
	 *
	 * when /Index is missing, a default value of [0 Size] is defined,
	 * where Size is the value of the /Size field. as in normal trailer
	 * dictionaries, it specifies the total size of the (entire)
	 * cross-reference table.
	 *
	 * when /W states a width of 0 for a field, that field is not present
	 * in the data and a default value should be used "if there is one".
	 * most notably, the first field determines the "type" of the entry,
	 * analogous to the 'n' and 'f' tags in classic xref sections; a width
	 * of 0 for the first field is specified to mean that every entry is of
	 * type 1 (= "n"). that type, in particular, specifies a default of 0
	 * for field 3 (generation). in fact, these are the only defaults
	 * defined by ISO 32000-1:2008 (PDF 1.7).
	 *
	 *   entry type  field no.  default value
	 *               1 (type)   1
	 *   1 ("n")     3 (gen.)   0
	 */

	/* Size (required) - total size of xref table */
	v = dictentry(dict, "Size");
	if (v == NULL || v->token_type != TT_SINT || v->sint < 1)
		return p_fail;
	Size = v->sint;

	/* W (required) - field widths for each xref entry */
	v = dictentry(dict, "W");
	if (v == NULL || v->token_type != TT_SEQUENCE)
		return p_fail;
	if ((Wn = v->seq->used) < 3)
		return p_fail;
	Wskip = 0;
	for (size_t i = 0; i < Wn; i++) {
		HTokenType tt = v->seq->elements[i]->token_type;
		int64_t w = v->seq->elements[i]->sint;

		if (tt != TT_SINT || w < 0)
			return p_fail;
		if (i < 3) {
			/* we can't take >64 bits and want to use size_t */
			if (w > 8 || (uint64_t)w > sizeof(size_t))
				return p_fail;
			W[i] = (size_t)w;
		} else {
			if ((uint64_t)w > SIZE_MAX - Wskip)
				return p_fail;	/* overflow */
			Wskip += w;
		}
	}
	if (Wskip > SIZE_MAX / 8)
		return p_fail;

	/*
	 * build the parser for one xref entry.
	 *
	 * in summary, the only sensible forms for /W are:
	 *
	 *   [t x y] with t,x,y > 0  full general form
	 *   [0 x y] with x,y > 0    only type-1 ("in use") entries
	 *   [0 x 0] with x > 0      only type-1 entries, only offsets
	 *
	 * however, though nonsensical, [t x 0] with t,x > 0 is not disallowed
	 * by the spec; as long as all entries are of type 1, the xref data can
	 * be interpreted without ambiguity.
	 *
	 * in fact, every nonsensical form is possible as long as there are 0
	 * entries.
	 *
	 * we realize this mess by just initializing the default parser to
	 * p_fail and and replacing the known cases afterwards.
	 */
	for (size_t i = 0; i < 3; i++) {
		if (W[i] == 0)
			p_field[i] = p_fail;	/* no known default */
		else
			p_field[i] = h_bits__m(mm__, W[i] * 8, false);
	}
	/* known default cases: */
	if (W[0] == 0)
		p_field[0] = p_return_1;	/* all type 1 */
	if (W[2] == 0) {
		p_field[2] = p_return_0;	/* all generation 0 */
		#if 0
		/* XXX
		 * i've seen a 0-width field 3 used with values of 1 (inuse)
		 * and 2 (objstm) in field 1, implying "objstm idx 0" for the
		 * latter case.
		 */
		if (W[0] > 0) {
			/* type field *must* be 1 */
			p_field[0] = h_attr_bool__m(mm__, p_field[0],
			    validate_eq_uint, (void *)1);
		}
		#endif
	}
	if (Wskip > 0)	// XXX h_skip does not work with CF, yet
		return p_fail;
	p_entry = h_sequence__m(mm__, p_field[0], p_field[1], p_field[2], NULL);
	p_entry = h_action__m(mm__, p_entry, act_xrstment, NULL);

	/* Index (optional) - subsections [base count ...] */
	v = dictentry(dict, "Index");
	if (v == NULL) {
		/* default: [0 Size] */
		p_subs = h_alloc(mm__, 2 * sizeof(HParser *));
		p_subs[0] = p_xrefsub__m(mm__, 0, Size, p_entry);
		p_subs[1] = NULL;
	} else if (v->token_type != TT_SEQUENCE) {
		return p_fail;
	} else {
		size_t nsubs = v->seq->used / 2;

		/* build a parser for each subsection */
		if (nsubs >= SIZE_MAX / sizeof(HParser *))
			return p_fail;
		p_subs = h_alloc(mm__, (nsubs + 1) * sizeof(HParser *));
		for (size_t i = 0; i < nsubs; i++) {
			HParsedToken *base = v->seq->elements[2 * i];
			HParsedToken *n = v->seq->elements[2 * i + 1];

			if (base->token_type != TT_SINT || base->sint < 0 ||
			    n->token_type != TT_SINT || n->sint < 0 ||
			    (uint64_t)n->sint > SIZE_MAX)
				return p_fail;

			p_subs[i] = p_xrefsub__m(mm__, base->sint, n->sint,
			    p_entry);
		}
		p_subs[nsubs] = NULL;
	}
	return h_sequence__ma(mm__, (void **)p_subs);
}



HParsedToken *
act_ostm(const HParseResult *p, void *u)
{
	assert(((HParsedToken *)u)->token_type == TT_SINT);
	size_t N = ((HParsedToken *)u)->sint;

	ObjStm   *ostrm = H_ALLOC(ObjStm);
	ostrm->numObjs  = N;
	ostrm->tok      = h_arena_malloc(p->arena, N* sizeof(Objref_T));
	ostrm->arena    = p->arena;

	for (int i=0; i<ostrm->numObjs; i++) {
		const HParsedToken *num = H_FIELD_TOKEN(0, 2*i);
		assert(num->token_type == TT_UINT);
		ostrm->tok[i].oid.nr  = H_CAST_UINT(num);
		ostrm->tok[i].oid.gen = 0;
		ostrm->tok[i].obj     = H_FIELD_TOKEN(1, i);
	}

//	const HCountedArray *indices = H_FIELD_SEQ(0);
//	const HCountedArray *ostrm   = H_FIELD_SEQ(1);

	const HParsedToken *tok = H_MAKE(ObjStm, ostrm);

	return (HParsedToken *)tok;
}

HParser *
p_objstm__m(HAllocator *mm__, const Dict *dict)
{
	const HParsedToken *v;
	size_t N;

	v = dictentry(dict, "N");
	if (v == NULL || v->token_type != TT_SINT || v->sint < 0 ||
	    (uint64_t)v->sint > SIZE_MAX) {
		log_message(7, "p_objstm__m: missing /N on object stream\n");
		return p_fail;
	}
	N = v->sint;

	HParser *wel_ws = h_sequence__m(mm__, p_wel, p_ws, NULL);
	HParser *idx = p_sepBy_n__m(mm__, p_npair, wel_ws, N);

	HParser *p_ostm = h_sequence__m(mm__, p_ws, idx, p_elemr, p_ws, NULL);
		// XXX plain indirect references not allowed in object streams
		// ISO 32000-2:2020, 7.5.7 "Object streams":
		// > An object in an object stream shall not consist solely of
		// > an object reference.
	HParser *ostm_p = h_action__m(mm__, p_ostm, act_ostm, (void *)v);

	return ostm_p;
		// XXX leading and trailing ws OK?

	// XXX consistency-check against /First, idx, /N
}

/*
 * This continuation is very similar to kstream, except that it does not
 * rely on /Length to consume the right amount of input. If /Length is
 * not present or indirect, it will operate on the entire rest of the input.
 * This is permissible, other than for general streams, because the XRef data
 * is always self-delimiting.
 *
 * x = (dict ...)
 */
HParser *
kxstream(HAllocator *mm__, const HParsedToken *x, void *env)
{
	struct Env *aux = env;
	HParsedToken *dict_t = H_INDEX_TOKEN(x, 0);
	Dict *dict = H_CAST(Dict, dict_t);
	const HParsedToken *v;
	HParser *bytes_p, *dict_p, *value_p;
	struct streamspec *spec;

	/* restrict bytes to Length if present (and not indirect) */
	v = dictentry(dict, "Length");
	if (v != NULL && v->token_type == TT_SINT && v->sint >= 0)
		bytes_p = p_take__m(mm__, v->sint, aux);
	else
		bytes_p = p_rest__m(mm__, aux);	// XXX consume the proper amount

	/* construct the parser for the stream data */
	spec = h_alloc(mm__, sizeof(struct streamspec));
	spec->dict = dict;
	spec->parser = p_xrefdata__m(mm__, dict);
	assert (spec->parser != NULL);

	dict_p  = p_return__m(mm__, dict_t);
	value_p = h_action__m(mm__, bytes_p, act_ks_value, spec);
		// XXX this should be h_bind, i think?! so failures in the
		// postordinate parser propagate...

	return h_sequence__m(mm__, dict_p, value_p, NULL);
}



/*
 * This helper implements the standard backwards parsing strategy to read all
 * cross-reference sections and trailer dictionaries, starting from the
 * 'startxref' offset found at the very end of the input.
 *
 * Allocates and returns an array of HParsedTokens, each containing the result
 * of a successful 'p_xref' parse. Sets the output parameter 'nxrefs' to the
 * number of elements.
 */
// XXX review changes to this function. see git -L /^parse_xrefs/,/^}/:pdf.c'
const char *infile = NULL;

void
parse_xrefs(struct Env *aux)
{
	const uint8_t *input = aux->input;
	size_t         sz    = aux->sz;
	HParseResult *res = NULL;
	const HParsedToken **xrefs = NULL;	/* empty result */
	const HParsedToken *tok = NULL;
	size_t n = 0, nfwd = 0;
	size_t offset = 0;
	bool processRoot = true;
	size_t maxObjNum = 0;
	Dict *trailer = NULL;

	// Initialize the environment variables
	aux->xrefs = xrefs;
	aux->nxrefs = n;



	/* search for the "startxref" section from the back of the file */
	HParser *p = h_left(p_startxref, h_end_p());
	for (size_t i = 0; i < sz; i++) {
		res = h_parse(p, input + sz - i, i);
		if (res != NULL)
			break;
	}
	if (res == NULL) {
		/* Severity should be SEV_DONTCARE, but we're simulating act_viol/VIOL() here */
		log_message(5, "VIOLATION[5]: startxref not found\n");
		return;
	}
	offset = H_INDEX_UINT(res->ast, 0);

	// XXX put many checks here into semantic validations in the parser...
	// e.g. offset bounds, /Prev type/value, loop detection

	// verify the offset recovered is bounded to be in the file
	if ( (offset <=0) || (offset >= aux->sz) ) {
		log_message(5, "VIOLATION[5]: Invalid xref table offset = %ld. Valid range <0, %ld>\n",
				offset, aux->sz);
		return;
	}


	for (;;) {
		//res = h_parse(p_xref, input + offset, sz - offset);
		p = NULL;
		p = h_right(h_seek(offset * 8, SEEK_SET), p_xref);	// XXX
		res = h_parse(p, input, sz);
		if (res == NULL || res->ast == NULL || H_INDEX_TOKEN(res->ast, 0) == NULL) {
			log_message(5, "VIOLATION[5]: error parsing xref section at "
			    "position %zu (%#zx)\n", offset, offset);
			return;
		}

		/* save this section in xrefs */
		if (n >= SIZE_MAX / sizeof(HParsedToken *))
			errx(2, "parse_xrefs: realloc: size would overflow");
		xrefs = realloc(xrefs, (n + 1) * sizeof(HParsedToken *));
		if (xrefs == NULL)
			err(2, "parse_xrefs");
		xrefs[n++] = res->ast;


		/* process the root */
		if (processRoot) {
			// Size is a required field in the trailer dictionary
			const HParsedToken *trailer_t = H_INDEX_TOKEN(res->ast, 1);
			trailer = H_CAST(Dict, trailer_t);
			const HParsedToken *size_t =  dictentry(trailer, "Size");
			if (size_t == NULL || size_t->token_type != TT_SINT) {
				log_message(5, "VIOLATION[5]: error parsing trailer section!"
							    "Missing or malformed -Size- field\n");
				return;
			}
			maxObjNum = H_CAST_SINT(size_t);

			processRoot = false;
		}


		/* look up the next offset (to the previous xref section) */
		tok = dictentry(H_INDEX(Dict, res->ast, 1), "Prev");
		if (tok == NULL)
			break;
		if (tok->token_type != TT_SINT) {
			log_message(7, "%s: /Prev not an integer\n", infile);
			break;
		}
		if (tok->sint < 0) {
			log_message(7, "%s: /Prev negative\n", infile);
			break;
		}

		/*
		 * validate the new offset. we don't want to get caught in a
		 * loop. the offsets should strictly decrease, unless the file
		 * is a "linearized" PDF. in that case there should be exactly
		 * one xref section at the beginning of the file that is
		 * allowed to point forward.
		 */
		if ((uint64_t)tok->sint >= offset)
			nfwd++;
		if (nfwd > 1) {
			log_message(7, "%s: /Prev pointer of xref section at "
			    "%zu (%#zx) points forward\n", infile, offset,
			    offset);
			break;
		}

		offset = (size_t)tok->sint;
	}



	/* Make sure we parsed a valid trailer section */
	if (! trailer) {
		log_message(7, "VIOLATION[7]: Invalid Trailer Section or Trailer Section not found\n");
		return;
	}



	// Process the document structure
	aux->xrefs = xrefs;
	aux->nxrefs = n;



	if (n > maxObjNum) {
		log_message(7, "%s: Number of xrefs found -%ld- "
				"Greater than specified /Size -%ld-.\n"
				"Ignoring objects numbered greater than -%ld-!\n",
				infile, n, maxObjNum, n);
		aux->nxrefs = maxObjNum;
	}


	// Process the trailer dictionary
	const HParsedToken *root = dictentry(trailer, "Root");
	assert(root->token_type == TT_Ref);
	parse_catalog(aux, root);

	return;
}



/*
 * ********************************************************************
 * End xref parsing
 * ********************************************************************
 */




/*
 * main program
 */

#include <stdio.h>
#include <inttypes.h>
#include <fcntl.h>	/* open() */
#include <unistd.h>	/* lseek(), getopt() */
#include <sys/mman.h>	/* mmap() */
#include <errno.h>

/* used with getopt() */
extern char *optarg;
extern int optind;

/* command line arguments */
const char *progname;
const char *xfile, *Xfile;
int qflag, vflag;

/* helper: parse 's' as a natural number. returns -1 on error. */
int
strtonat(const char *s)
{
	char *p;
	long l;

	errno = 0;
	l = strtol(s, &p, 10);
	if (errno != 0)
		return -1;
	if (p == s || *p != '\0')	/* no parse */
		return -1;
	if (l < 0 || l > INT_MAX)	/* out of range */
		return -1;
	return l;
}

void
usage(void)
{
	fprintf(stderr, "usage: %s [-qsv] [-x out.txt] file\n", progname);
	exit(2);
}

int
main(int argc, char *argv[])
{
	struct Env aux;
	HParseResult *res = NULL;
	const uint8_t *input;
	off_t sz;
	int fd;
	int ch, n;

	/* command line handling */
	progname = argv[0];
	while ((ch = getopt(argc, argv, "L:qsS:vx:X:")) != -1) {
		switch(ch) {
		case 'L':	/* set "logging" threshold */
			n = strtonat(optarg);
			if (n == -1)
				errx(2, "-L: invalid argument");
			log_level = n;
			break;
		case 'q':	/* query/quiet mode */
			qflag = 1;
			break;
		case 's':	/* strict mode */
			strictness = 1;
			break;
		case 'S':	/* set severity threshold */
			n = strtonat(optarg);
			if (n == -1)
				errx(2, "-S: invalid argument");
			strictness = n;
			break;
		case 'v':	/* verbose mode */
			vflag = 1;
			break;
		case 'x':
			xfile = optarg;
			break;
		case 'X':
			Xfile = optarg;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
		usage();
	infile = argv[0];

	/* configure output verbosity */
	log_level = strictness + 1;		/* any errors by default */
	if (qflag)
		log_level = SEV_DONTCARE;	/* only program errors */
	else if (vflag)
		log_level = 0;			/* all messages */
	if (qflag && vflag)			/* "quiet verbose mode" */
		log_level = strictness + 1;	/* error messages back on */

	/* mmap the input file */
	fd = open(infile, O_RDONLY);
	if (fd == -1)
		err(2, "%s", infile);
	sz = lseek(fd, 0, SEEK_END);
	if (sz == -1)
		err(2, "lseek");
	assert(sz >= 0);
	input = mmap(NULL, sz?sz:1, PROT_READ, MAP_PRIVATE, fd, 0);
	if (input == MAP_FAILED)
		err(2, "mmap");

	/* build parsers */
	aux = (struct Env){infile, input, sz};
	init_parser(&aux);
	init_LZW_parser();

	/* parse all cross-reference sections and trailer dictionaries */
	parse_xrefs(&aux);

	/* run the main parser */
	res = h_parse(p_pdf, input, sz);
	if (!res) {
		if (!qflag) {
			fprintf(stderr, "%s: no parse\n", infile);

			/* help us find the error */
			res = h_parse(p_pdfdbg, input, sz);
			if (res) {
				int64_t pos = res->bit_length / 8;
				fprintf(stderr, "%s: error after position"
				    " %" PRId64 " (%#" PRIx64 ")\n",
				    infile, pos, pos);
				//h_pprintln(stderr, res->ast);	// XXX debug
			}
		}

		return 1;
	}

	/* print result */
	if (!qflag)
		h_pprintln(stdout, res->ast);

	/* Save the extracted text */
	if (aux.ntextobjs > 0)
		text_extract(&aux, xfile, Xfile);

	return 0;
}
